# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Any, Dict, List, Optional, Tuple, Union
from warnings import warn

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd

from jaxley.modules import Branch, Cell, Compartment, Network
from jaxley.utils.cell_utils import v_interp
from jaxley.utils.morph_attributes import (
    morph_attrs_from_xyzr,
    split_xyzr_into_equal_length_segments,
)

########################################################################################
###################################### HELPERS #########################################
########################################################################################


def _is_leaf(graph: nx.Graph, node: Any) -> bool:
    """Return whether or not a node in an undirected graph is a leaf."""
    # For a leaf node, the degree is is 1. For a single compartment neuron it is 0.
    return graph.degree(node) < 2


def _is_branching(graph: nx.Graph, node: Any) -> bool:
    """Return whether an undirected graph is branching at a particular node."""
    return graph.degree(node) > 2


def _has_same_id(
    graph: nx.Graph,
    node_i: Any,
    node_j: Any,
    relevant_type_ids: List[int],
):
    """Return whether two nodes in a graph have the same value for the `id` attribute.

    Args:
        relevant_type_ids: All type ids that are not in this list will be ignored for
            tracing the morphology. This means that branches which have multiple type
            ids (which are not in `relevant_type_ids`) will be considered as one branch.
    """
    if (
        graph.nodes[node_i]["id"] not in relevant_type_ids
        and graph.nodes[node_j]["id"] not in relevant_type_ids
    ):
        return True
    else:
        return graph.nodes[node_i]["id"] == graph.nodes[node_j]["id"]


def _get_soma_idxs(graph: nx.Graph):
    """Return all SWC nodes which have id=1, i.e. which are labelled as soma."""
    return [i for i, n in nx.get_node_attributes(graph, "id").items() if n == 1]


def _unpack(d: Dict, keys: list[str]) -> List:
    """Return all values of a dictionary whose key is in `keys`."""
    return [d[k] for k in keys]


def _branch_e2n(branch_edges):
    """Return all nodes given the edges within a branch.

    E.g. `branch_edges = [(0, 1), (1, 2)]` -> `[0, 1, 2]`.
    """
    array = np.concatenate(branch_edges)
    return array[np.sort(np.unique(array, return_index=True)[1])]


def _branch_n2e(branch_nodes):
    """Return all edges given the nodes within a branch.

    E.g. `branch_nodes = [0, 1, 2]` -> [(0, 1), (1, 2)]`"""
    return [e for e in zip(branch_nodes[:-1], branch_nodes[1:])]


def _find_root(G: nx.Graph):
    """Return a possible root for tracing the graph.

    Roots are nodes which have degree = 0."""
    roots = [n for n in sorted(G.nodes) if _is_leaf(G, n)]
    return roots[0]


########################################################################################
################################### BUILD SWC GRAPH ####################################
########################################################################################


def to_swc_graph(fname: str, num_lines: int = None) -> nx.DiGraph:
    """Read a SWC file and return a SWC graph via networkX.

    The graph is read such that each entry in the swc file becomes a graph node
    with the column attributes (id, x, y, z, r). Then each node is connected to its
    designated parent via an edge. A "type" attribute is added to the graph to identify
    its processing stage for subsequent steps.

    Args:
        fname: Path to the swc file.
        num_lines: Number of lines to read from the file. If None, all lines are read.

    Returns:
        A networkx graph of the traced morphology in the swc file. It has attributes:
        nodes: {'id': 1, 'x': 0.0, 'y': 0.0, 'z': 0.0, 'r': 1.0, 'p': -1}
        edges: {'l': 1.0}

    Example usage
    ^^^^^^^^^^^^^

    ::

        from jaxley.io.graph to_swc_graph
        swc_graph = to_swc_graph("path_to_swc.swc")
    """
    i_id_xyzr_p = np.loadtxt(fname)[:num_lines]

    graph = nx.DiGraph()
    graph.add_nodes_from(
        (
            (int(i), {"id": int(id), "x": x, "y": y, "z": z, "r": r, "p": int(p)})
            for i, id, x, y, z, r, p in i_id_xyzr_p
        )
    )
    graph.add_edges_from([(p, i) for p, i in i_id_xyzr_p[:, [-1, 0]] if p != -1])
    return _add_missing_graph_attrs(graph)


def _add_missing_graph_attrs(graph: nx.Graph) -> nx.Graph:
    """Add missing attributes to the graph nodes and edges.

    The following attributes are added to the graph:
    - id: int (default: 0)
    - x, y, z: float (default: NaN)
    - r: float (default: 1)
    - l: float (default: 1)
    - p: int (default: 0)

    Args:
        graph: A networkx graph.

    Returns:
        The graph with the added attributes."""
    available_keys = graph.nodes[1].keys()
    defaults = {
        "id": 0,
        "x": float("nan"),
        "y": float("nan"),
        "z": float("nan"),
        "r": 1,
        "p": 0,
    }
    # add defaults if not present
    for key in set(defaults.keys()).difference(available_keys):
        nx.set_node_attributes(graph, defaults[key], key)

    graph = _add_edge_lengths(graph)
    edge_lens = nx.get_edge_attributes(graph, "l")
    if np.isnan(list(edge_lens.values())[0]):
        nx.set_edge_attributes(graph, 1, "l")

    return graph


def _add_edge_lengths(graph: nx.Graph, min_len: float = 1e-5) -> nx.DiGraph:
    """Add edge lengths to graph.edges based on the xyz coordinates of graph.nodes."""
    xyz = lambda i: np.array(_unpack(graph.nodes[i], "xyz"))
    for i, j in graph.edges:
        d_ij = (
            np.sqrt(((xyz(i) - xyz(j)) ** 2).sum())
            if i != j
            else 2 * graph.nodes[i]["r"]
        )
        # min_len ensures that 2 nodes cannot lie on top of each other
        # this is important for the compartmentalization
        graph.edges[i, j]["l"] = d_ij if d_ij >= min_len else min_len
    return graph


########################################################################################
############################## BUILD COMPARTMENT GRAPH #################################
########################################################################################


def build_compartment_graph(
    swc_graph: nx.DiGraph,
    ncomp: int,
    root: Optional[int] = None,
    min_radius: Optional[float] = None,
    max_len=None,
    ignore_swc_tracing_interruptions=True,
    relevant_type_ids: Optional[List[int]] = None,
) -> nx.DiGraph:
    """Return a networkX graph that indicates the compartment structure.

    Build a new graph made up of compartments in every branch. These compartments are
    spaced at equidistant points along the branch. Node attributes, like radius are
    linearly interpolated along its length.

    Example: 4 compartments | edges = - | nodes = o | comp_nodes = x
    o-----------o----------o---o---o---o--------o
    o-------x---o----x-----o--xo---o---ox-------o

    This function returns a _directed_ graph. The graph is directed only because every
    compartment tracks the xyzr coordinates of the associated SWC file. These xyzr
    coordinates are ordered by the order of the traversal of the swc_graph. In later
    methods (e.g. build_solve_graph), we traverse the `comp_graph` and mostly ignore
    the directionality of the edges, but we only use the directionality to reverse the
    xyzr coordinates if necessary.

    Args:
        swc_graph: Graph generated by `to_swc_graph()`.
        ncomp: How many compartments per branch to insert.
        root: The root branch from which to start tracing the nodes. This defines the
            branch indices.
        min_radius: Minimal radius for each compartment.
        max_len: Maximal length for each branch. Longer branches are split into
            separate branches.
        ignore_swc_tracing_interruptions: If `False`, it this function automatically
            starts a new branch when a section is traced with interruptions.
        relevant_type_ids: All type ids that are not in this list will be ignored for
            tracing the morphology. This means that branches which have multiple type
            ids (which are not in `relevant_type_ids`) will be considered as one branch.
            If `None`, we default to `[1, 2, 3, 4]`.

    Returns:
        Directed graph made up of compartments.

        Compartment nodes have attributes (example):
        {'x': 2.0,
        'y': 0.0,
        'z': 0.0,
        'branch_index': 2,
        'comp_index': 2,
        'type': 'comp',
        'xyzr': array([[0., 0., 0., 1.], [1., 0., 0., 1.]]),
        'groups': ['soma'],
        'radius': 1.0,
        'length': 4.0,
        'cell_index': 0}

        Between-compartment nodes have attributes:
        {'x': 0.0,
        'y': 0.0,
        'z': 0.0,
        'p': -1,
        'type': 'branchpoint',
        'groups': ['soma'],
        'radius': 1.0,
        'length': 0.0,
        'cell_index': 0}

        Edges have attributes: {}

    Example usage
    ^^^^^^^^^^^^^

    ::

        from jaxley.io.graph to_swc_graph, build_compartment_graph
        swc_graph = to_swc_graph("path_to_swc.swc")
        comp_graph = build_compartment_graph(swc_graph, ncomp=1)
    """
    swc_graph, branch_edge_indices, all_type_ids, soma_ignore_inds = _trace_branches(
        swc_graph,
        root=root,
        max_len=max_len,
        ignore_swc_tracing_interruptions=ignore_swc_tracing_interruptions,
        relevant_type_ids=relevant_type_ids,
    )
    comp_offset = 0

    # See docstring for why `comp_graph` a directed graph.
    comp_graph = nx.DiGraph()

    # Get branchpoints.
    branchpoint_inds = [
        n for n in swc_graph.nodes if swc_graph.nodes[n]["type"] == "branchpoint"
    ]
    comp_graph.add_nodes_from((n, swc_graph.nodes[n]) for n in branchpoint_inds)
    nx.set_node_attributes(comp_graph, 0.0, "l")

    for branch_index, branch_edge_inds in enumerate(branch_edge_indices):
        path_lens = np.cumsum(np.concatenate([[0], branch_edge_inds[:, 2]]))

        # [:, :2] because `branch_edge_inds` contains `(start_node, end_note, length)`.
        branch_nodes = _branch_e2n(branch_edge_inds[:, :2])

        # `branch_data` is a pd.DataFrame which contains all SWC nodes of the current
        # branch.
        branch_data = pd.DataFrame([swc_graph.nodes[i] for i in branch_nodes])
        branch_data["node_index"] = branch_nodes
        branch_data["l"] = path_lens

        # errors="ignore" because user-defined graphs might not have the `p` attribute.
        branch_data = branch_data.drop(columns=["type", "p"], errors="ignore")

        # fix id and r bleed over from neighboring neurites of a different type
        if branch_data.loc[0, "id"] != branch_data.loc[1, "id"]:
            branch_data.loc[0, ["r", "id"]] = branch_data.loc[1, ["r", "id"]]

        xyzr = branch_data[["x", "y", "z", "r"]].to_numpy()

        # `soma_ignore_inds` tracks all node indices which are part of a single-point-
        # soma or of a somatic branchpoint. In these cases, the somatic SWC node
        # is _not_ considered to be part of the dendrite. Here, we delete somatic
        # SWC node from the xyzr of the dendrite.
        if branch_edge_inds[0, 0] in soma_ignore_inds:
            xyzr = xyzr[1:]

        # Here, we split xyzr into compartments.
        xyzr_per_comp = split_xyzr_into_equal_length_segments(xyzr, ncomp)
        morph_attrs = np.asarray(
            [morph_attrs_from_xyzr(xyzr, min_radius, ncomp) for xyzr in xyzr_per_comp]
        )

        branch_len = branch_data["l"].max()
        if branch_len < 1e-8:
            warn(
                "Found a branch with length 0. To avoid NaN while integrating the "
                "ODE, we capped this length to 0.1 um. The underlying cause for the "
                "branch with length 0 is likely a strange SWC file. The "
                "most common reason for this is that the SWC contains a soma "
                "traced by a single point, and a dendrite that connects to the soma "
                "has no further child nodes."
            )
            branch_len = 0.1
        comp_len = branch_len / ncomp
        locs = np.linspace(comp_len / 2, branch_len - comp_len / 2, ncomp)

        # New branch_nodes is a pd.DataFrame which contains all ncomp compartments that
        # make up the branch.
        new_branch_nodes = v_interp(locs, branch_data["l"].values, branch_data.values)
        new_branch_nodes = pd.DataFrame(
            np.array(new_branch_nodes.T), columns=branch_data.columns
        )
        new_branch_nodes["id"] = all_type_ids[
            branch_index
        ]  # new_branch_nodes["id"].astype(int)
        new_branch_nodes["l"] = comp_len
        new_branch_nodes["branch_index"] = branch_index
        new_branch_nodes["comp_index"] = comp_offset + np.arange(ncomp)
        num_nodes = max(comp_graph.nodes) + 1 if comp_graph.nodes else 0
        new_branch_nodes["node_index"] = num_nodes + np.arange(ncomp)
        new_branch_nodes["type"] = "comp"
        comp_offset += ncomp
        new_branch_nodes["xyzr"] = xyzr_per_comp
        new_branch_nodes["r"] = morph_attrs[:, 0]
        new_branch_nodes["area"] = morph_attrs[:, 1]
        new_branch_nodes["volume"] = morph_attrs[:, 2]
        new_branch_nodes["resistive_load_in"] = morph_attrs[:, 3]
        new_branch_nodes["resistive_load_out"] = morph_attrs[:, 4]

        # Add the compartments as nodes to the new `comp_graph`.
        new_branch_nodes = new_branch_nodes.set_index("node_index")
        comp_graph.add_nodes_from(new_branch_nodes.to_dict(orient="index").items())

        # Add the edges between compartments within a node.
        new_branch_edges = _branch_n2e(new_branch_nodes.index)
        comp_graph.add_edges_from(new_branch_edges)

        # Add edge from compartment ending & beginning to branchpoint.
        pre_branch_node = branch_edge_inds[0, 0]
        post_branch_node = branch_edge_inds[-1, 1]
        if (
            pre_branch_node in comp_graph.nodes
            and comp_graph.nodes[pre_branch_node]["type"] == "branchpoint"
        ):
            comp_graph.add_edge(pre_branch_node, new_branch_nodes.index[0])
        if (
            post_branch_node in comp_graph.nodes
            and comp_graph.nodes[post_branch_node]["type"] == "branchpoint"
        ):
            comp_graph.add_edge(new_branch_nodes.index[-1], post_branch_node)

    # Delete `comp_index`,... from branchpoint nodes.
    for node, attrs in comp_graph.nodes(data=True):
        if comp_graph.nodes[node]["type"] == "branchpoint":
            for key in ["comp_index", "branch_index", "xyzr"]:
                if key in attrs:
                    del comp_graph.nodes[node][key]

    # Rename all nodes to have the compartment index as node name _if they are
    # compartments_ and otherwise to have a larger indices as node names
    # (for branchpoints).
    mapping = {}
    branchpoint_index = 0
    for n, attrs in comp_graph.nodes(data=True):
        if "comp_index" in attrs:
            mapping[n] = attrs["comp_index"]
        else:
            mapping[n] = f"n{branchpoint_index}"
            branchpoint_index += 1
    comp_graph = nx.relabel_nodes(comp_graph, mapping, copy=True)
    comp_graph = _set_branchpoint_indices(comp_graph)

    min_radius = min_radius if min_radius else 0.0

    # Rename attributes.
    # Description of SWC file format:
    # http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html
    group_ids = {0: "undefined", 1: "soma", 2: "axon", 3: "basal", 4: "apical"}
    min_radius = min_radius if min_radius else 0.0
    clip_radius = lambda r: max(r, min_radius) if min_radius else r
    for n in comp_graph.nodes:
        type_id = comp_graph.nodes[n].pop("id")
        comp_graph.nodes[n]["groups"] = [group_ids.get(type_id, f"custom{type_id}")]
        comp_graph.nodes[n]["radius"] = clip_radius(comp_graph.nodes[n].pop("r"))
        comp_graph.nodes[n]["length"] = comp_graph.nodes[n].pop("l")

    # Here, we assume that the SWC file was all from a single neuron. However, this is
    # fine anyways because we assert that it is a tree.
    nx.set_node_attributes(comp_graph, 0, "cell_index")
    return comp_graph


def _trace_branches(
    swc_graph: nx.DiGraph,
    root: Optional[int] = None,
    max_len: Optional[float] = None,
    ignore_swc_tracing_interruptions: bool = True,
    relevant_type_ids: Optional[List[int]] = None,
) -> List[np.ndarray]:
    """Get all uninterrupted paths in a graph (i.e. branches).

    The graph is traversed depth-first starting from the source node, which is the only
    node with in_degree 0 (raises in case multiple are found). Note: Traversal order
    can be changed, by reversing the edge directions, i.e. to start traversal from a
    leaf node set: [source -> ... -> leaf] to [leaf -> ... -> source].

    The `graph` is modified in two ways:
    - For single-point somata we modify some things. In addition, we make the graph
    undirected.
    - Add `branchpoint: bool` as a node attribute.

    Args:
        graph: A networkx graph.
        ignore_swc_tracing_interruptions: Whether to ignore discontinuities in the swc
            tracing order. If False, this will result in split branches at these points.
        relevant_type_ids: All type ids that are not in this list will be ignored for
            tracing the morphology. This means that branches which have multiple type
            ids (which are not in `relevant_type_ids`) will be considered as one branch.
            If `None`, we default to `[1, 2, 3, 4]`.

    Returns:
        A list of linear paths in the graph. Each path is represented as an array of
        edges.
    """
    if relevant_type_ids is None:
        relevant_type_ids = [1, 2, 3, 4]

    # `soma_ignore_inds` tracks the node indices of single-point-somata or of
    # somatic branchpoints.
    soma_ignore_inds = []

    # Handle special case of a single soma node.
    soma_idxs = _get_soma_idxs(swc_graph)
    if len(soma_idxs) == 1:
        soma = soma_idxs[0]

        for i, j in (*swc_graph.in_edges(soma), *swc_graph.out_edges(soma)):
            swc_graph.edges[i, j]["l"] = 0

        # For single-point somatata, we set
        # l = 2*r ensures A_cylinder = 2*pi*r*l = 4*pi*r^2 = A_sphere.
        # Here, we add another compartment such that there exists an `edge` between
        # that new compartment and the single-point-soma of appropriate length.
        swc_graph.nodes[1]["p"] = 0
        for n in swc_graph.nodes:
            swc_graph.nodes[n]["p"] += 1
        swc_graph.add_node(0, **swc_graph.nodes[1])
        swc_graph.nodes[0]["p"] = -1
        swc_graph.add_edge(0, soma, l=2 * swc_graph.nodes[soma]["r"])
        swc_graph = nx.relabel_nodes(swc_graph, {i: i + 1 for i in swc_graph.nodes})
        soma_ignore_inds.append(1)
        soma_ignore_inds.append(soma + 1)

    undir_swc_graph = swc_graph.to_undirected()

    # Handle somatic branchpoints. A somatic branchpoint is a branchpoint at which at
    # least two connecting branches are somatic. In that case (and in the case of a
    # single-point soma), non-somatic branches are assumed to start from their first
    # traced point, not from the soma.
    for node in soma_idxs:
        somatic_neighbors = [
            n
            for n in undir_swc_graph.neighbors(node)
            if undir_swc_graph.nodes[n]["id"] == 1
        ]
        if len(somatic_neighbors) > 1:
            for i, j in undir_swc_graph.edges(node):
                if undir_swc_graph.nodes[j]["id"] != 1:
                    soma_ignore_inds.append(i)
                    undir_swc_graph.edges[i, j]["l"] = 0

    branches, current_branch, all_type_ids = [], [], []

    # Traverse the SWC graph and identify which nodes belong to one branch. This builds
    # a list `branches: List` where each elements is a `np.array` of shape (N, 3).
    # The `3` are `swc_parent, swc_node, length` of all SWC edges within a branch.
    root = root if root else _find_root(undir_swc_graph)

    # We first set the type of all SWC nodes to be "spurious". Later on, we change the
    # type of branchpoints to `branchpoint`.
    nx.set_node_attributes(undir_swc_graph, "spurious", "type")
    # `sort_neighbors=lambda x: sorted(x)` to first handle edges with a low node index.
    for i, j in nx.dfs_edges(undir_swc_graph, root, sort_neighbors=lambda x: sorted(x)):
        current_edge = (i, j)
        current_len = undir_swc_graph.edges[current_edge]["l"]
        current_type_id = undir_swc_graph.nodes[i]["id"]
        current_branch += [(i, j, current_len)]
        if _is_leaf(undir_swc_graph, j):
            # If the SWC leaf node has a different type_id than the its SWC
            # predecessor, then we explicitly must add both branches here.
            if not _has_same_id(undir_swc_graph, i, j, relevant_type_ids):
                # Add the branch that goes up until the last edge.
                branches.append(current_branch[:-1])
                all_type_ids.append(current_type_id)

                # Add the branch made up of just the last edge.
                branches.append(current_branch[-1:])
                all_type_ids.append(undir_swc_graph.nodes[j]["id"])

                current_branch = []
            else:
                branches.append(current_branch)
                all_type_ids.append(current_type_id)
                current_branch = []

        elif _is_branching(undir_swc_graph, j):
            branches.append(current_branch)
            all_type_ids.append(current_type_id)
            current_branch = []

        # Start new branch if ids differ.
        elif not _has_same_id(undir_swc_graph, i, j, relevant_type_ids):
            # Consider the SWC graph:
            # 1  1  1  2  2  2 (number indicates type_id)
            if not swc_graph.has_edge(i, j) and swc_graph.has_edge(j, i):
                branches.append(current_branch)
                all_type_ids.append(current_type_id)
                current_branch = []
            else:
                branches.append(current_branch[:-1])
                all_type_ids.append(current_type_id)
                current_branch = [current_branch[-1]]

    branch_edges = []
    type_inds = []
    for i, p in enumerate(branches):
        if len(p) > 0:
            branch_edges.append(np.array(p))
            type_inds.append(all_type_ids[i])

    if max_len:
        edge_lens = nx.get_edge_attributes(undir_swc_graph, "l")
        additional_branchpoints, branch_edges, type_inds = _split_branches(
            branch_edges, type_inds, edge_lens, max_len
        )
        for b in additional_branchpoints:
            undir_swc_graph.nodes[b]["type"] = "branchpoint"

    # Label nodes in the swc_graph as `branchpoint`.
    #
    # The very first branch will start at a tip, not a branchpoint (therefore:
    # `branch_edges[1:]`). All other branches start at a branch-point. To get all
    # branchpoints, we loop over all branches but the first one and get their first
    # traced node.
    for b in branch_edges[1:]:
        tip = b[0, 0]
        undir_swc_graph.nodes[tip]["type"] = "branchpoint"

    if not ignore_swc_tracing_interruptions:
        # Ignore added index by default; only relevant in case it was added.
        additional_branchpoints, branch_edges, type_inds = (
            _split_branches_if_swc_nodes_were_traced_with_interruption(
                undir_swc_graph, branch_edges, type_inds
            )
        )
        for b in additional_branchpoints:
            undir_swc_graph.nodes[b]["type"] = "branchpoint"

    return undir_swc_graph, branch_edges, type_inds, soma_ignore_inds


def _split_branches(
    branches: List[np.ndarray], type_inds, edge_lens: Dict, max_len: int = 1000
) -> List[np.ndarray]:
    """Split branches into approximately equally long sections <= max_len.

    Args:
        branches: List of branches represented as arrays of edges.
        edge_lens: Dict for length of each edge in the graph.
        max_len: Maximum length of a branch section. If a branch exceeds this length,
            it is split into equal parts.

    Returns:
        A list of branches, where each branch is split into sections of
        length <= max_len.
    """
    # TODO: split branches into exactly equally long sections
    edge_lens.update({(j, i): l for (i, j), l in edge_lens.items()})
    additional_branchpoints, new_branches, new_type_inds = [], [], []
    for branch, type_ind in zip(branches, type_inds):
        cum_branch_len = np.cumsum([edge_lens[i, j] for i, j, _ in branch])

        k = cum_branch_len // max_len
        split_branch = [branch[np.where(np.array(k) == kk)[0]] for kk in np.unique(k)]
        new_branches += split_branch
        new_type_inds += [type_ind] * len(split_branch)

        # Introduce additional branchpoints.
        #
        # Ignore the first one (via [1:]) because that node is anyways labeled as a
        # branchpoint.
        additional_branchpoints += [int(branch[0, 0]) for branch in split_branch[1:]]

    return additional_branchpoints, new_branches, new_type_inds


def _split_branches_if_swc_nodes_were_traced_with_interruption(
    graph: nx.Graph, branches: List[np.ndarray], type_inds: List[int]
) -> List[np.ndarray]:
    """Simulate swc trace errors in the branches.

    Both NEURON and Jaxley's hand coded swc reader introduce breaks in the trace
    if the same neurite was traced in disconnected pieces. Since `swc_to_graph` is
    agnostic to the order of the tracing, it does not suffer from this issue. Hence,
    to artificially force this behaviour (to compare to the other parsers), this
    function can be used to simulate these errors. See
    `_find_swc_tracing_interruptions` for how to identify these points in the graph.

    Args:
        graph: A networkx graph of a traced morphology.
        branches: List of branches represented as arrays of edges.

    Returns:
        A list of branches with simulated trace errors.
    """
    node_idxs = _find_swc_tracing_interruptions(graph)
    for node_idx in node_idxs:
        # Get index of the first branch in which `node_idx` appears.
        # [:, :2] to get rid of the length. `p` is of shape `(num_nodes_in_branch, 3)`,
        # where the 3 are `parent, node, length`.
        branch_idx = next(i for i, p in enumerate(branches) if node_idx in p[:, :2])
        b4, branch, after = (
            branches[:branch_idx],
            branches[branch_idx],
            branches[branch_idx + 1 :],
        )
        type_b4, type_val, type_after = (
            type_inds[:branch_idx],
            type_inds[branch_idx],
            type_inds[branch_idx + 1 :],
        )
        # [:, :2] to get rid of the length. `p` is of shape `(num_nodes_in_branch, 3)`,
        # where the 3 are `parent, node, length`.
        break_idx = np.where(branch[:, :2] == node_idx)[0][1]
        # insert artificial break into branch
        branches = b4 + [branch[:break_idx], branch[break_idx:]] + after
        type_inds = type_b4 + [type_val, type_val] + type_after
    return node_idxs, branches, type_inds


def _find_swc_tracing_interruptions(graph: nx.Graph) -> np.ndarray:
    """Identify discontinuities in the swc tracing order.

    Some swc files contain artefacts, where tracing of the same neurite was done
    in disconnected pieces. Both NEURON and Jaxley's hand coded swc reader introduce
    a break in the trace at these points, since they parse the file in order. This
    leads to split branches, which should be one. This function identifies these
    points in the graph.

    Example swc file:
    # branch 1
    1 1 0.0 0.0 0.0 1.0 -1
    2 1 1.0 0.0 0.0 1.0 1
    3 1 2.0 0.0 0.0 1.0 2
    # branch 2
    4 2 3.0 1.0 0.0 1.0 3
    5 2 4.0 2.0 0.0 1.0 4
    # branch 3
    6 3 3.0 -1.0 0.0 1.0 3
    7 3 4.0 -2.0 0.0 1.0 6
    8 3 5.0 -3.0 0.0 1.0 7
    # amend branch 2
    9 4 5.0 3.0 0.0 1.0 5

    Args:
        graph: graph tracing of swc file (from `swc_to_graph`).

    Returns:
        An array of node indices where tracing is discontinuous.
    """
    interrupted_nodes = []
    for n in graph.nodes:
        parent = graph.nodes[n]["p"]
        # Parent should be previous node.
        if parent > -1 and parent != n - 1:
            node_is_no_branchpoint = graph.nodes[n]["type"] != "branchpoint"
            parent_is_no_branchpoint = graph.nodes[parent]["type"] != "branchpoint"
            if node_is_no_branchpoint and parent_is_no_branchpoint:
                interrupted_nodes.append(parent)

    return interrupted_nodes


def _set_branchpoint_indices(jaxley_graph: nx.DiGraph) -> nx.DiGraph:
    """Return a graph whose branchpoint indices match those of a `jx.Module`.

    Here, we ensure that the branchpoints are enumerated in the same way in the
    module as they are in the graph. The ordering is by the branch_index of the
    parent branch of a branchpoint.
    """
    predecessor_branch_inds = []
    branchpoints = []
    max_comp_index = 0
    for node in jaxley_graph.nodes:
        if jaxley_graph.nodes[node]["type"] == "branchpoint":
            predecessor = list(jaxley_graph.predecessors(node))[0]
            predecessor_branch_inds.append(
                jaxley_graph.nodes[predecessor]["branch_index"]
            )
            branchpoints.append(node)
        else:
            if jaxley_graph.nodes[node]["comp_index"] > max_comp_index:
                max_comp_index = jaxley_graph.nodes[node]["comp_index"]
    sorting = np.argsort(predecessor_branch_inds)
    branchpoints_in_corrected_order = np.asarray(branchpoints)[sorting]
    mapping = {
        k: max_comp_index + i + 1 for i, k in enumerate(branchpoints_in_corrected_order)
    }
    return nx.relabel_nodes(jaxley_graph, mapping)


########################################################################################
################################ BUILD SOLVE GRAPH #####################################
########################################################################################


def _set_comp_and_branch_index(
    comp_graph: nx.DiGraph,
    root: Optional[int] = None,
) -> nx.DiGraph:
    """Given a compartment graph, return a comp_graph with new comp and branch index.

    The returned comp and branch index are the ones used also in the resulting
    jx.Cell, and they define the solve order for `jaxley.stone` solvers.

    Args:
        comp_graph: Compartment graph returned by `build_compartment_graph`.
        root: The root node to traverse the graph for the solve order.

    Returns:
        A directed graph indicating the solve order. The graph does no longer contain
        branchpoints. Nodes contain the following attributes (example):
        ```{'x': 0.0,
        'y': 3.0,
        'z': 0.0,
        'branch_index': 0,
        'comp_index': 0,
        'type': 'comp',
        'xyzr': array([[0., 4., 0., 1.], [0., 3., 0., 1.]]),
        'groups': ['axon'],
        'radius': 1.0,
        'length': 2.0,
        'cell_index': 0}```
    """
    undirected_comp_graph = comp_graph.to_undirected()
    root = root if root else _find_root(undirected_comp_graph)

    # Directed graph to store the traversal
    solve_graph = nx.DiGraph()

    # Copy all global attributes over to the solve_graph.
    for key in comp_graph.graph.keys():
        solve_graph.graph[key] = comp_graph.graph[key]
    solve_graph.add_nodes_from(undirected_comp_graph.nodes(data=True))
    solve_graph.nodes[root]["comp_index"] = 0
    solve_graph.nodes[root]["branch_index"] = 0

    comp_index = 1
    branch_index = 0
    node_inds_in_which_to_flip_xyzr = []

    # Traverse the graph for the solve order.
    # `sort_neighbors=lambda x: sorted(x)` to first handle nodes with lower node index.
    node_mapping = {root: 0}
    for i, j in nx.dfs_edges(
        undirected_comp_graph, root, sort_neighbors=lambda x: sorted(x)
    ):
        solve_graph.add_edge(i, j)
        solve_graph.nodes[j]["branch_index"] = branch_index
        if solve_graph.nodes[j]["type"] == "comp":
            solve_graph.nodes[j]["comp_index"] = comp_index
            node_mapping[j] = comp_index

        if _is_leaf(undirected_comp_graph, j):
            branch_index += 1

        # Increase the branch counter if a branchpoint is encountered.
        elif undirected_comp_graph.nodes[j]["type"] == "branchpoint":
            branch_index += 1

        # Increase the counter for the compartment index only if the node was a
        # compartment (branchpoints are skipped).
        if solve_graph.nodes[j]["type"] == "comp":
            comp_index += 1

        # The `xyzr` attribute of all compartment nodes is ordered in the order in
        # which the SWC file was traversed. If we now traverse a compartment from
        # another direction (because the solve order is not the same as the SWC trace
        # order), then we have to flip the `xyzr` coordinates.
        if not comp_graph.has_edge(i, j) and comp_graph.has_edge(j, i):
            node_inds_in_which_to_flip_xyzr.append(i)
            node_inds_in_which_to_flip_xyzr.append(j)

    unique_nodes_to_flip = list(set(node_inds_in_which_to_flip_xyzr))
    for n in unique_nodes_to_flip:
        # Branchpoint nodes do not have the xyzr property.
        if "xyzr" in solve_graph.nodes[n].keys():
            solve_graph.nodes[n]["xyzr"] = solve_graph.nodes[n]["xyzr"][::-1]

    solve_graph = nx.relabel_nodes(solve_graph, node_mapping)
    solve_graph = _set_branchpoint_indices(solve_graph)
    return solve_graph


def _remove_branch_points_at_tips(comp_graph: nx.DiGraph) -> nx.DiGraph:
    """Delete branch points at tips.

    These only occur if the user was editing the morphology."""
    nodes_to_keep = []
    for node in comp_graph.nodes:
        degree = comp_graph.in_degree(node) + comp_graph.out_degree(node)
        if degree > 1 or comp_graph.nodes[node]["type"] == "comp":
            nodes_to_keep.append(node)
    return comp_graph.subgraph(nodes_to_keep).copy()


def _remove_branch_points(solve_graph: nx.DiGraph) -> nx.DiGraph:
    """Remove branch points and label edges as `inter_branch` or `intra_branch`."""

    # Copy the graph because, otherwise, its input gets modified.
    solve_graph = solve_graph.copy()

    # All connections which either have no `type` label or which are not labelled as
    # synapses are labelled as `intra_branch` for. `inter_branch` connections are
    # handled in the loop below.
    for edge_ind in solve_graph.edges:
        edge = solve_graph.edges[edge_ind]
        if "type" not in edge or edge["type"] != "synapse":
            edge["type"] = "intra_branch"

    # Replace branch points with direct connections between the compartments.
    for node in list(solve_graph.nodes):
        if solve_graph.nodes[node].get("type") == "branchpoint":
            parents = list(solve_graph.predecessors(node))
            children = list(solve_graph.successors(node))
            for v in children:
                for u in parents:
                    solve_graph.add_edge(u, v, type="inter_branch")
            solve_graph.remove_node(node)

    return solve_graph


def _add_meta_data(solve_graph: nx.DiGraph) -> nx.DiGraph:
    """Return a graph with some attributes renamed for Jaxley compatibility.

    The returned graph is what we call the `Jaxley` graph as it is the exact graph
    that can be fully read and written by `Jaxley`.

    Args:
        solve_graph: Directed graph build by `build_solve_graph()`.
        min_radius: If not None, clips the radiuses of all compartments.

    Returns:
        Directed graph which indicates the solve order (and the order of compartments)
        within a Jaxley module. Each node has the following attributes (example, when
        it was read from an SWC file):
        ```{'x': 1.16,
        'y': 5.16,
        'z': 1.0,
        'type': 'comp',
        'xyzr': array([[2., 6., 1., 1.], [1., 5., 1., 1.]]),
        'comp_index': 0,
        'branch_index': 0,
        'cell_index': 0,
        'groups': ['axon'],
        'radius': 1.0,
        'length': 1.18}```
    """
    branch_inds = nx.get_node_attributes(solve_graph, "branch_index")
    branch_edge_df = pd.DataFrame(branch_inds.items(), columns=["node", "branch_index"])
    all_xyzr = []
    # The `.apply(list).sort_index().items()` ensures that the `branch_index` is sorted.
    # This is important because we expect all_xyzr[n] to correspond to the n-th
    # branch.
    for branch_index, nodes_in_branch in (
        branch_edge_df.groupby("branch_index")["node"].apply(list).sort_index().items()
    ):
        nodes_in_branch = sorted(nodes_in_branch)
        xyzr = [solve_graph.nodes[n]["xyzr"] for n in nodes_in_branch]
        xyzr = np.concatenate(xyzr)
        all_xyzr.append(xyzr)

    solve_graph.graph["xyzr"] = all_xyzr

    # It will typically have a `type` if the graph was exported from a module. If it
    # does not, then we make it a cell (which assumes that the read SWC was a cell, but
    # we assert for this anyways (by making sure that it is a `nx.tree`).
    if "type" not in solve_graph.graph:
        solve_graph.graph["type"] = "cell"

    return solve_graph  # This is now a `jaxley_graph`.


########################################################################################
################################ MODULE FROM GRAPH #####################################
########################################################################################


def from_graph(
    comp_graph: nx.DiGraph,
    assign_groups: bool = True,
    solve_root: Optional[int] = None,
    traverse_for_solve_order: bool = True,
):
    """Return a Jaxley module from a compartmentalized networkX graph.

    This method is currently limited to graphs that have the same number of
    compartments in each branch. If this is not the case then the method will raise
    an `AssertionError`.

    Args:
        comp_graph: The compartment graph built with `build_compartment_graph()` or
            with `to_graph()`.
        assign_groups: Whether to assign groups to the nodes.
        solve_root: The root node to traverse the graph for identifying the solve
            order.
        traverse_for_solve_order: Whether to traverse the graph for identifying the
            solve order. Should only be set to `False` if you are confident that the
            `comp_graph` is in a form in which it can be solved (i.e. its branch
            indices, compartment indices, and node names are correct). Typically, this
            is the case only if you exported a module to a `comp_graph` via `to_graph`,
            did not modify the graph, and now re-import it as a module with
            `from_graph`.

    Return:
        A `jx.Module` representing the graph.

    Example usage
    ^^^^^^^^^^^^^

    ::

        from jaxley.io.graph import to_swc_graph, build_compartment_graph, from_graph
        swc_graph = to_swc_graph("path_to_swc.swc")
        comp_graph = build_compartment_graph(swc_graph, ncomp=1)
        cell = from_graph(comp_graph)
    """
    comp_graph = _remove_branch_points_at_tips(comp_graph)

    # If the graph is based on a custom-built `jx.Module` (e.g., parents=[-1, 0, 0, 1]),
    # and we did not modify the exported graph, then we might not want to traverse
    # the graph again because this would change the ordering of the branches.
    if traverse_for_solve_order:
        comp_graph = _set_comp_and_branch_index(comp_graph, root=solve_root)

    solve_graph = _remove_branch_points(comp_graph)
    solve_graph = _add_meta_data(solve_graph)
    module = _build_module(solve_graph, assign_groups=assign_groups)
    return module


def _build_module(
    solve_graph: nx.DiGraph,
    assign_groups: bool = True,
):
    # nodes and edges
    node_df = pd.DataFrame(
        [d for i, d in solve_graph.nodes(data=True)], index=solve_graph.nodes
    ).sort_index()

    node_df = node_df.drop(columns=["xyzr", "type"])
    edge_type = nx.get_edge_attributes(solve_graph, "type")
    synapse_edges = pd.DataFrame(
        [
            {
                "pre_index": i,
                "post_index": j,
                **solve_graph.edges[i, j],
            }
            for (i, j), t in edge_type.items()
            if t == "synapse"
        ]
    )

    # branches
    branch_of_node = lambda i: solve_graph.nodes[i]["branch_index"]
    branch_edges_df = pd.DataFrame(
        [
            (branch_of_node(i), branch_of_node(j))
            for (i, j), t in edge_type.items()
            if t == "inter_branch"
        ],
        columns=["parent_branch_index", "child_branch_index"],
    )

    # drop special attrs from nodes and ignore error if col does not exist
    # x,y,z can be re-computed from xyzr if needed
    optional_attrs = [
        "recordings",
        "externals",
        "external_inds",
        "trainable",
    ]
    node_df = node_df.drop(columns=optional_attrs, errors="ignore")

    # synapses
    synapse_edges = synapse_edges.drop(["l", "type"], axis=1, errors="ignore")
    synapse_edges = synapse_edges.rename({"syn_type": "type"}, axis=1)
    synapse_edges.rename({"edge_index": "global_edge_index"}, axis=1, inplace=True)

    # build module
    acc_parents = []
    parent_branch_inds = branch_edges_df.set_index("child_branch_index").sort_index()[
        "parent_branch_index"
    ]
    assert np.std(node_df.groupby("branch_index").size().to_numpy()) < 1e-8, (
        "`from_graph()` does not support a varying number of compartments in each "
        "branch."
    )
    for branch_inds in node_df.groupby("cell_index")["branch_index"].unique():
        root_branch_idx = branch_inds[0]
        parents = parent_branch_inds.loc[branch_inds[1:]] - root_branch_idx
        acc_parents.append([-1] + parents.tolist())

    # TODO: support inhom ncomps
    module = _build_module_scaffold(
        node_df, solve_graph.graph["type"], acc_parents, solve_graph.graph["xyzr"]
    )

    # set global attributes of module. `xyzr` is passed here again, although it has
    # already been passed to `_build_module_scaffold`. `jx.Cell` requires xyzr at
    # __init__()`, all other modules do not.
    for k, v in solve_graph.graph.items():
        if k not in ["type"]:
            setattr(module, k, v)

    if assign_groups and "groups" in node_df.columns:
        groups = node_df.pop("groups").explode()
        groups = (
            pd.DataFrame(groups)
            .groupby("groups")
            .apply(lambda x: x.index.values, include_groups=False)
            .to_dict()
        )
        for group_name, group_inds in groups.items():
            module.select(nodes=group_inds).add_to_group(group_name)

    node_df.columns = [
        "global_" + col if "local" not in col and "index" in col else col
        for col in node_df.columns
    ]
    # set column-wise. preserves cols not in df.
    module.nodes[node_df.columns] = node_df
    module.edges = synapse_edges if not synapse_edges.empty else module.edges

    # add all the extra attrs
    module.membrane_current_names = [c.current_name for c in module.channels]
    module.synapse_names = [s._name for s in module.synapses]

    return module


def _build_module_scaffold(
    idxs: pd.DataFrame,
    return_type: Optional[str] = None,
    parent_branches: Optional[List[np.ndarray]] = None,
    xyzr: List[np.ndarray] = [],
) -> Union[Network, Cell, Branch, Compartment]:
    """Builds a skeleton module from a DataFrame of indices.

    This is useful for instantiating a module that can be filled with data later.

    Args:
        idxs: DataFrame containing cell_index, branch_index, comp_index, i.e.
            Module.nodes or View.view.
        return_type: Type of module to return. If None, the type is inferred from the
            number of unique values in the indices. I.e. only 1 unique cell_index
                and 1 unique branch_index -> return_type = "jx.Branch".

    Returns:
        A skeleton module with the correct number of compartments, branches, cells, or
        networks."""
    return_types = ["compartment", "branch", "cell", "network"]
    build_cache = {k: [] for k in return_types}

    if return_type is None:  # infer return type from idxs
        return_type = _infer_module_type_from_inds(idxs)

    comp = Compartment()
    build_cache["compartment"] = [comp]

    if return_type in return_types[1:]:
        ncomps = idxs["branch_index"].value_counts().iloc[0]
        branch = Branch([comp for _ in range(ncomps)])
        build_cache["branch"] = [branch]

    if return_type in return_types[2:]:
        branch_counter = 0
        for cell_id, cell_groups in idxs.groupby("cell_index"):
            num_branches = cell_groups["branch_index"].nunique()
            default_parents = np.arange(num_branches) - 1  # ignores morphology
            parents = (
                default_parents if parent_branches is None else parent_branches[cell_id]
            )
            cell = Cell(
                [branch] * num_branches,
                parents,
                xyzr=xyzr[branch_counter : branch_counter + num_branches],
            )
            build_cache["cell"].append(cell)
            branch_counter += num_branches

    if return_type in return_types[3:]:
        build_cache["network"] = [Network(build_cache["cell"])]

    module = build_cache[return_type][0]
    build_cache.clear()
    return module


def _infer_module_type_from_inds(idxs: pd.DataFrame) -> str:
    """Return type (comp, branch, cell, ...) given dataframe of indices."""
    nuniques = idxs[["cell_index", "branch_index", "comp_index"]].nunique()
    nuniques.index = ["cell", "branch", "compartment"]
    nuniques = pd.concat([pd.Series({"network": 1}), nuniques])
    return_type = nuniques.loc[nuniques == 1].index[-1]
    return return_type


########################################################################################
#################################### VISUALIZATION #####################################
########################################################################################


def vis_compartment_graph(
    comp_graph: nx.DiGraph,
    ax=None,
    font_size: float = 7.0,
    node_size: float = 150.0,
    arrowsize: float = 10.0,
    comp_color: str = "r",
    branchpoint_color: str = "orange",
):
    """Visualize a compartment graph.

    Args:
        comp_graph: A compartment graph generated with `build_compartment_graph()` or
            with `to_graph()`.
        ax: Matplotlib axis.
        font_size: The fontsize for the node names.
        node_size: The size of each node.
        arrowsize: The size of the arrow.
        comp_color: The color of the compartments.
        branchpoint_color: The color of the compartments.

    Example usage
    ^^^^^^^^^^^^^

    ::

        from jaxley.io.graph import to_graph, vis_compartment_graph
        cell = jx.read_swc("path_to_swc.swc", ncomp=1)
        comp_graph = to_graph(cell)
        vis_compartment_graph(comp_graph)
    """
    color_map = []
    for n in comp_graph.nodes:
        if comp_graph.nodes[n].get("type") == "comp":
            new_col = comp_color
        elif comp_graph.nodes[n].get("type") == "branchpoint":
            new_col = branchpoint_color
        color_map.append(new_col)

    pos = {k: (v["x"], v["y"]) for k, v in comp_graph.nodes.items()}
    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(4, 4))

    nx.draw(
        comp_graph,
        pos=pos,
        with_labels=True,
        font_size=font_size,
        node_size=node_size,
        ax=ax,
        node_color=color_map,
        arrowsize=arrowsize,
    )


########################################################################################
########################### UTILITIES FOR MODIFYING GRAPHS #############################
########################################################################################


def connect_graphs(
    graph1: nx.DiGraph,
    graph2: nx.DiGraph,
    node1: Union[str, int],
    node2: Union[str, int],
) -> nx.DiGraph:
    """Return a new graph that connects two comp_graphs at particular nodes."""
    # For each `group` and `channel` in `graph1`, ensure that it is `False` in `graph2`
    # (if it does not exist).
    graph2 = _assign_false_for_group_and_channel(graph1, graph2)
    graph1 = _assign_false_for_group_and_channel(graph2, graph1)

    # Move graph2 such that it smoothly connects to graph1.
    for i, key in enumerate(["x", "y", "z"]):
        coord1 = _infer_coord(graph1, node1, key)
        coord2 = _infer_coord(graph2, node2, key)
        offset = coord1 - coord2
        for node in graph2.nodes:
            graph2.nodes[node][key] += offset
            if graph2.nodes[node]["type"] == "comp":
                graph2.nodes[node]["xyzr"][:, i] += offset

    # Rename the nodes of graph2.
    offset_comps = max([n for n in graph1.nodes])
    mapping = {}
    for n in sorted(graph2.nodes):
        new_index = n + offset_comps + 1
        mapping[n] = new_index

    graph2 = nx.relabel_nodes(graph2, mapping)
    node2 = mapping[node2]

    # Combine the graph1 and graph2 into one graph.
    combined_graph = nx.compose(graph1, graph2)

    # By default, nx.compose uses the graph-level attributes from graph2. We want that
    # if a graph level attribute is a list, then the graph level attribute of the
    # combined_graph should be a concatenation. This is done below. It ensures that,
    # e.g., channels from both modules are concatenated.
    for key in set(graph1.graph) | set(graph2.graph):
        val1 = graph1.graph.get(key)
        val2 = graph2.graph.get(key)

        if isinstance(val1, list) and isinstance(val2, list):
            combined_graph.graph[key] = val1 + val2
        elif key in graph2.graph:
            combined_graph.graph[key] = val2
        else:
            combined_graph.graph[key] = val1

    # Add edges between graph1 and graph2 to connect them. The code below differentiates
    # three cases: comp->comp, branchpoint->comp (or comp->branchpoint) and
    # branchpoint->branchpoint.
    type1 = combined_graph.nodes[node1]["type"]
    type2 = combined_graph.nodes[node2]["type"]
    offset_branchpoints = max([n for n in combined_graph.nodes])
    if type1 == "comp" and type2 == "comp":
        # If both nodes are compartments, then we insert a new branchpoint.
        #
        # Search for the first node labelled as `type=branchpoint`. Once we have found
        # such a node, we `break`.
        for node in combined_graph.nodes:
            if combined_graph.nodes[node]["type"] == "branchpoint":
                new_attrs = combined_graph.nodes(data=True)[node].copy()
                break
        # Set the xyz coordinates of the new node.
        for key in ["x", "y", "z"]:
            comp1_xyz = combined_graph.nodes[node1][key]
            comp2_xyz = combined_graph.nodes[node2][key]
            new_attrs[key] = 0.5 * (comp1_xyz + comp2_xyz)
        new_node_index = offset_branchpoints + 1
        combined_graph.add_node(new_node_index, **new_attrs)
        combined_graph.add_edge(node1, new_node_index)
        combined_graph.add_edge(new_node_index, node2)
    elif type1 == "comp" or type2 == "comp":
        # If one of the nodes is a compartment and the other one is not, then
        # we just connect.
        combined_graph.add_edge(node1, node2)
    else:
        # Delete branchpoint in second graph. Connect all nodes that it connected to
        # to the first branchpoint.
        for i in combined_graph.predecessors(node2):
            combined_graph.add_edge(i, node1)
        for i in combined_graph.successors(node2):
            combined_graph.add_edge(node1, i)
        combined_graph.remove_node(node2)

    # Add the graph attributes (which are not carried over when doing `compose`)
    group_names = set()
    if "group_names" in graph1.graph.keys():
        group_names = group_names | set(graph1.graph["group_names"])
    if "group_names" in graph2.graph.keys():
        group_names = group_names | set(graph2.graph["group_names"])
    combined_graph.graph["group_names"] = list(group_names)

    # Relabel the compartments. Before the code below, compartments induced by `graph2`
    # have a higher node index than branchpoints of `graph1`. Below, we fix this.
    mapping = {}
    counter = 0
    for n in sorted(combined_graph.nodes):
        if combined_graph.nodes[n]["type"] == "comp":
            mapping[n] = counter
            counter += 1
    for n in sorted(combined_graph.nodes):
        if combined_graph.nodes[n]["type"] == "branchpoint":
            mapping[n] = counter
            counter += 1
    return nx.relabel_nodes(combined_graph, mapping)


def _assign_false_for_group_and_channel(
    graph1: nx.DiGraph, graph2: nx.DiGraph
) -> nx.DiGraph:
    """For any group and channel in graph1.graph, set False in nodes of graph2.

    Args:
        graph1: The graph in which to check for `group_names`.
        graph2: The graph whose `group_names` to update.
    """
    channel_names = [channel.__class__.__name__ for channel in graph1.graph["channels"]]
    if "group_names" in graph1.graph.keys():
        for group in graph1.graph["group_names"] + channel_names:
            for node in graph2.nodes:
                if group not in graph2.nodes[node].keys():
                    graph2.nodes[node][group] = False
    return graph2


def _infer_coord(comp_graph: nx.DiGraph, node: Union[str, int], key: str) -> float:
    """Return the coordinate of a node in a comp_graph.

    Args:
        key: Either of x, y, z.
    """
    i = {"x": 0, "y": 1, "z": 2}[key]
    if comp_graph.nodes[node]["type"] == "branchpoint":
        coordinate = comp_graph.nodes[node][key]
    else:
        if comp_graph.in_degree(node) == 1:
            coordinate = comp_graph.nodes[node]["xyzr"][-1, i]
        else:
            coordinate = comp_graph.nodes[node]["xyzr"][0, i]
    return coordinate



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import matplotlib.pyplot as plt
import numpy as np


def network_cols(num_neurons_in_layers: list):
    fn_order = [
        grey_shades,
        purple_shades,
        blue_shades,
        green_shades,
        red_shades,
        orange_shades,
    ]
    cols = []
    for num_in_layer, col_builder in zip(num_neurons_in_layers, fn_order):
        cols_in_layer = col_builder(num_in_layer)
        cols += cols_in_layer
    return cols


def blue_shades(n):
    """Return n shades of blue."""
    return plt.get_cmap("Blues")(np.linspace(1.0, 0.2, n)).tolist()


def purple_shades(n):
    """Return n shades of purple."""
    return plt.get_cmap("Purples")(np.linspace(1.0, 0.2, n)).tolist()


def red_shades(n):
    """Return n shades of red."""
    return plt.get_cmap("Reds")(np.linspace(1.0, 0.2, n)).tolist()


def green_shades(n):
    """Return n shades of green."""
    return plt.get_cmap("Greens")(np.linspace(1.0, 0.2, n)).tolist()


def grey_shades(n):
    """Return n shades of grey."""
    return plt.get_cmap("Greys")(np.linspace(1.0, 0.2, n)).tolist()


def orange_shades(n):
    """Return n shades of orange."""
    return plt.get_cmap("Oranges")(np.linspace(1.0, 0.2, n)).tolist()



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax.numpy as jnp
import numpy as np
import pytest
from jax import jit

import jaxley as jx
import jaxley.optimize.transforms as jt
from jaxley.optimize.transforms import ParamTransform


def test_joint_inverse():
    # test forward(inverse(x))=x
    tf_dict = [
        {"param_array_1": jt.SigmoidTransform(-2, 2)},
        {"param_array_2": jt.SoftplusTransform(2)},
        {"param_array_3": jt.NegSoftplusTransform(-2)},
    ]

    params = [
        {"param_array_1": jnp.asarray(np.linspace(-1, 1, 4))},
        {"param_array_2": jnp.asarray(np.linspace(-4, 1, 24))},
        {"param_array_3": jnp.asarray(np.linspace(-1, 4, 4))},
    ]

    tf = ParamTransform(tf_dict)
    forward = tf.forward(params)
    inverse = tf.inverse(forward)

    assert np.allclose(
        inverse[0]["param_array_1"], params[0]["param_array_1"]
    ), "SigmoidTransform forward, inverse failed."
    assert np.allclose(
        inverse[1]["param_array_2"], params[1]["param_array_2"]
    ), "SoftplusTransform forward, inverse failed."
    assert np.allclose(
        inverse[2]["param_array_3"], params[2]["param_array_3"]
    ), "NegSoftplusTransform forward, inverse failed."


def test_bounds():
    # test if forward maps parameters into bounds
    lowers = {"param_array_1": -2, "param_array_2": None, "param_array_3": -2}
    uppers = {"param_array_1": 2, "param_array_2": 2, "param_array_3": None}

    tf_dict = [
        {"param_array_1": jt.SigmoidTransform(-2, 2)},
        {"param_array_2": jt.NegSoftplusTransform(2)},
        {"param_array_3": jt.SoftplusTransform(-2)},
    ]

    params = [
        {"param_array_1": jnp.asarray(np.linspace(-10, 10, 4))},
        {"param_array_2": jnp.asarray(np.linspace(-10, 10, 4))},
        {"param_array_3": jnp.asarray(np.linspace(-10, 10, 4))},
    ]

    tf = ParamTransform(tf_dict)
    forward = tf.forward(params)

    assert all(
        forward[0]["param_array_1"] > lowers["param_array_1"]
    ), "SigmoidTransform failed to match lower bound."
    assert all(
        forward[0]["param_array_1"] < uppers["param_array_1"]
    ), "SigmoidTransform failed to match upper bound."
    assert all(
        forward[1]["param_array_2"] < uppers["param_array_2"]
    ), "SoftplusTransform failed to match lower bound."
    assert all(
        forward[2]["param_array_3"] > lowers["param_array_3"]
    ), "NegSoftplusTransform failed to match lower bound."


@pytest.mark.parametrize(
    "transform",
    [
        jt.SigmoidTransform(-2, 2),
        jt.SoftplusTransform(2),
        jt.NegSoftplusTransform(2),
        jt.AffineTransform(1.0, 1.0),
        jt.CustomTransform(lambda x: x, lambda x: x),
        jt.ChainTransform([jt.SigmoidTransform(-2, 2), jt.SoftplusTransform(2)]),
    ],
)
def test_jit(transform):
    # test jit-compilation:
    tf_dict = [{"param_array_1": transform}]

    params = [{"param_array_1": jnp.asarray(np.linspace(-1, 1, 4))}]

    tf = ParamTransform(tf_dict)

    @jit
    def test_jit(params):
        return tf.inverse(params)

    _ = test_jit(params)


@pytest.mark.parametrize(
    "transform",
    [
        jt.SigmoidTransform(-2, 2),
        jt.SoftplusTransform(2),
        jt.NegSoftplusTransform(2),
        jt.AffineTransform(1.0, 1.0),
        jt.CustomTransform(lambda x: x, lambda x: x),
        jt.ChainTransform([jt.SigmoidTransform(-2, 2), jt.SoftplusTransform(2)]),
        jt.MaskedTransform(
            jnp.array([True, False, False, True]), jt.SigmoidTransform(-2, 2)
        ),
    ],
)
def test_correct(transform):
    # Test correctness on "standard" PyTree
    tf_dict = [{"param_array_1": transform}]

    params = [{"param_array_1": jnp.asarray(np.linspace(-1, 1, 4))}]

    tf = ParamTransform(tf_dict)

    forward = tf.forward(params)
    inverse = tf.inverse(forward)

    assert np.allclose(
        inverse[0]["param_array_1"], params[0]["param_array_1"]
    ), f"{transform} forward, inverse failed."

    # Test correctness plain Array
    for shape in [(4,), (4, 1), (4, 4)]:
        tf_dict = transform
        tf = ParamTransform(tf_dict)
        x = jnp.ones(shape)
        y = tf.forward(x)
        x_inv = tf.inverse(y)

        assert np.allclose(
            x, x_inv
        ), f"{transform} forward, inverse failed on non PyTree."


@pytest.mark.parametrize(
    "transform",
    [jt.SigmoidTransform(-2, 2), jt.SoftplusTransform(2), jt.NegSoftplusTransform(2)],
)
def test_user_api(transform, SimpleCell):
    cell = SimpleCell(3, 2)

    cell.branch("all").make_trainable("radius")
    cell.branch(2).make_trainable("radius")
    cell.branch(1).make_trainable("length")
    cell.branch(0).make_trainable("radius")
    cell.branch("all").comp("all").make_trainable("axial_resistivity")
    cell.make_trainable("capacitance")

    params = cell.get_parameters()

    # We scale it to something small as axial_resistivity is large
    # and then the transform becomes numerically uninvertible.
    t = jt.ChainTransform([jt.AffineTransform(1e-3, 0.0), transform])

    tf_dict = [{list(k.keys())[0]: t} for k in params]
    tf = ParamTransform(tf_dict)

    forward = tf.forward(params)
    reverse = tf.inverse(forward)

    flat_params, _ = jax.tree_util.tree_flatten(params)
    flat_reverse, _ = jax.tree_util.tree_flatten(reverse)
    assert all(
        [np.allclose(a, b) for a, b in zip(flat_params, flat_reverse)]
    ), f"{transform} forward, inverse failed."



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>
from typing import List

import jax.numpy as jnp
import numpy as np
import pandas as pd
from jax import Array, vmap
from jax.typing import ArrayLike

from jaxley.utils.misc_utils import cumsum_leading_zero


def equal_segments(branch_property: list, ncomp_per_branch: int):
    """Generates segments where some property is the same in each segment.

    Args:
        branch_property: List of values of the property in each branch. Should have
            `len(branch_property) == num_branches`.
    """
    assert isinstance(branch_property, list), "branch_property must be a list."
    return jnp.asarray([branch_property] * ncomp_per_branch).T


def linear_segments(
    initial_val: float, endpoint_vals: list, parents: ArrayLike, ncomp_per_branch: int
):
    """Generates segments where some property is linearly interpolated.

    Args:
        initial_val: The value at the tip of the soma.
        endpoint_vals: The value at the endpoints of each branch.
    """
    branch_property = endpoint_vals + [initial_val]
    num_branches = len(parents)
    # Compute radiuses by linear interpolation.
    endpoint_radiuses = jnp.asarray(branch_property)

    def compute_rad(branch_ind, loc):
        start = endpoint_radiuses[parents[branch_ind]]
        end = endpoint_radiuses[branch_ind]
        return (end - start) * loc + start

    branch_inds_of_each_comp = jnp.tile(jnp.arange(num_branches), ncomp_per_branch)
    locs_of_each_comp = jnp.linspace(1, 0, ncomp_per_branch).repeat(num_branches)
    rad_of_each_comp = compute_rad(branch_inds_of_each_comp, locs_of_each_comp)

    return jnp.reshape(rad_of_each_comp, (ncomp_per_branch, num_branches)).T


def merge_cells(
    cumsum_num_branches: List[int],
    cumsum_num_branchpoints: List[int],
    arrs: List[List[np.ndarray]],
    exclude_first: bool = True,
) -> np.ndarray:
    """
    Build full list of which branches are solved in which iteration.

    From the branching pattern of single cells, this "merges" them into a single
    ordering of branches.

    Args:
        cumsum_num_branches: cumulative number of branches. E.g., for three cells with
            10, 15, and 5 branches respectively, this will should be a list containing
            `[0, 10, 25, 30]`.
        arrs: A list of a list of arrays that should be merged.
        exclude_first: If `True`, the first element of each list in `arrs` will remain
            unchanged. Useful if a `-1` (which indicates "no parent") entry should not
            be changed.

    Returns:
        A list of arrays which contain the branch indices that are computed at each
        level (i.e., iteration).
    """
    ps = []
    for i, att in enumerate(arrs):
        p = att
        if exclude_first:
            raise NotImplementedError
            p = [p[0]] + [p_in_level + cumsum_num_branches[i] for p_in_level in p[1:]]
        else:
            p = [
                p_in_level
                + np.asarray([cumsum_num_branches[i], cumsum_num_branchpoints[i]])
                for p_in_level in p
            ]
        ps.append(p)

    max_len = max([len(att) for att in arrs])
    combined_parents_in_level = []
    for i in range(max_len):
        current_ps = []
        for p in ps:
            if len(p) > i:
                current_ps.append(p[i])
        combined_parents_in_level.append(np.concatenate(current_ps))

    return combined_parents_in_level


def compute_levels(parents):
    levels = np.zeros_like(parents)

    for i, p in enumerate(parents):
        if p == -1:
            levels[i] = 0
        else:
            levels[i] = levels[p] + 1
    return levels


def compute_children_in_level(
    levels: np.ndarray, children_row_and_col: np.ndarray
) -> List[np.ndarray]:
    num_branches = len(levels)
    children_in_each_level = []
    for l in range(1, np.max(levels) + 1):
        children_in_current_level = []
        for b in range(num_branches):
            if levels[b] == l:
                children_in_current_level.append(children_row_and_col[b - 1])
        children_in_current_level = np.asarray(children_in_current_level)
        children_in_each_level.append(children_in_current_level)
    return children_in_each_level


def _compute_num_children(parents):
    num_branches = len(parents)
    num_children = []
    for b in range(num_branches):
        n = np.sum(np.asarray(parents) == b)
        num_children.append(n)
    return num_children


def _compute_index_of_child(parents):
    """For every branch, it returns the how many-eth child of its parent it is.

    Example:
    ```
    parents = [-1, 0, 0, 1, 1, 1]
    _compute_index_of_child(parents) -> [-1, 0, 1, 0, 1, 2]
    ```
    """
    num_branches = len(parents)
    current_num_children_for_each_branch = np.zeros((num_branches,), np.dtype("int"))
    index_of_child = [-1]
    for b in range(1, num_branches):
        index_of_child.append(current_num_children_for_each_branch[parents[b]])
        current_num_children_for_each_branch[parents[b]] += 1
    return index_of_child


def compute_children_indices(parents) -> list[Array]:
    """Return all children indices of every branch.

    Example:
    ```
    parents = [-1, 0, 0]
    compute_children_indices(parents) -> [[1, 2], [], []]
    ```
    """
    num_branches = len(parents)
    child_indices = []
    for b in range(num_branches):
        child_indices.append(np.where(parents == b)[0])
    return child_indices


def get_num_neighbours(
    num_children: ArrayLike,
    ncomp_per_branch: int,
    num_branches: int,
):
    """
    Number of neighbours of each compartment.
    """
    num_neighbours = 2 * jnp.ones((num_branches * ncomp_per_branch))
    num_neighbours = num_neighbours.at[ncomp_per_branch - 1].set(1.0)
    num_neighbours = num_neighbours.at[jnp.arange(num_branches) * ncomp_per_branch].set(
        num_children + 1.0
    )
    return num_neighbours


def local_index_of_loc(
    loc: float, global_branch_ind: int, ncomp_per_branch: int
) -> int:
    """Returns the local index of a comp given a loc [0, 1] and the index of a branch.

    This is used because we specify locations such as synapses as a value between 0 and
    1. We have to convert this onto a discrete segment here.

    Args:
        branch_ind: Index of the branch.
        loc: Location (in [0, 1]) along that branch.
        ncomp_per_branch: Number of segments of each branch.

    Returns:
        The local index of the compartment.
    """
    ncomp = ncomp_per_branch[global_branch_ind]  # only for convenience.
    possible_locs = np.linspace(0.5 / ncomp, 1 - 0.5 / ncomp, ncomp)
    ind_along_branch = np.argmin(np.abs(possible_locs - loc))
    return ind_along_branch


def loc_of_index(global_comp_index, global_branch_index, ncomp_per_branch):
    """Return location corresponding to global compartment index."""
    cumsum_ncomp = cumsum_leading_zero(ncomp_per_branch)
    index = global_comp_index - cumsum_ncomp[global_branch_index]
    ncomp = ncomp_per_branch[global_branch_index]
    return (0.5 + index) / ncomp


def remap_to_consecutive(arr):
    """Maps an array of integers to an array of consecutive integers.

    E.g. `[0, 0, 1, 4, 4, 6, 6] -> [0, 0, 1, 2, 2, 3, 3]`
    """
    _, inverse_indices = jnp.unique(arr, return_inverse=True)
    return inverse_indices


v_interp = vmap(jnp.interp, in_axes=(None, None, 1))


def interpolate_xyzr(loc: float, coords: np.ndarray):
    """Perform a linear interpolation between xyz-coordinates.

    Args:
        loc: The location in [0,1] along the branch.
        coords: Array containing the reconstructed xyzr points of the branch.

    Return:
        Interpolated xyz coordinate at `loc`, shape `(3,).
    """
    dl = np.sqrt(np.sum(np.diff(coords[:, :3], axis=0) ** 2, axis=1))
    pathlens = np.insert(np.cumsum(dl), 0, 0)  # cumulative length of sections
    norm_pathlens = pathlens / np.maximum(1e-8, pathlens[-1])  # norm lengths to [0,1].

    return v_interp(loc, norm_pathlens, coords)


def params_to_pstate(
    params: list[dict[str, ArrayLike]],
    indices_set_by_trainables: list[ArrayLike],
):
    """Make outputs `get_parameters()` conform with outputs of `.data_set()`.

    `make_trainable()` followed by `params=get_parameters()` does not return indices
    because these indices would also be differentiated by `jax.grad` (as soon as
    the `params` are passed to `def simulate(params)`. Therefore, in `jx.integrate`,
    we run the function to add indices to the dict. The outputs of `params_to_pstate`
    are of the same shape as the outputs of `.data_set()`."""
    return [
        {"key": list(p.keys())[0], "val": list(p.values())[0], "indices": i}
        for p, i in zip(params, indices_set_by_trainables)
    ]


def convert_point_process_to_distributed(current: ArrayLike, area: ArrayLike) -> Array:
    """Convert current point process (nA) to distributed current (uA/cm2).

    This function gets called for synapses and for external stimuli.

    Args:
        current: Current in `nA`.
        area: Membrane surface area radius in `um^2`.

    Return:
        Current in `uA/cm2`.
    """
    current /= area  # nA / um^2
    return current * 100_000  # Convert (nA / um^2) to (uA / cm^2)


def query_channel_states_and_params(d, keys, idcs):
    """Get dict with subset of keys and values from d.

    This is used to restrict a dict where every item contains __all__ states to only
    the ones that are relevant for the channel. E.g.

    ```states = {'eCa': Array([ 0.,  0., nan]}```

    will be
    ```states = {'eCa': Array([ 0.,  0.]}```

    Only loops over necessary keys, as opposed to looping over `d.items()`."""
    return dict(zip(keys, (v[idcs] for v in map(d.get, keys))))


def compute_children_and_parents(
    branch_edges: pd.DataFrame,
) -> tuple[Array, Array, Array, int]:
    """Build indices used during `._init_morph_custom_spsolve()."""
    par_inds = branch_edges["parent_branch_index"].to_numpy()
    child_inds = branch_edges["child_branch_index"].to_numpy()
    child_belongs_to_branchpoint = remap_to_consecutive(par_inds)
    par_inds = np.unique(par_inds)
    return par_inds, child_inds, child_belongs_to_branchpoint


def _get_comp_edges_in_view(
    comp_edges: pd.DataFrame, incl_comps, comp_edge_condition: str
):
    """Return the `_comp_edges` within a current `View`.

    Args:
        base_comp_edges: `comp_edges` to be filtered.
        incl_comps: Sequence of compartments that are within the current `View`.
        comp_edge_condition: Either of
            {`source_and_sink`, `source_or_sink`, `endpoint`, `startpoint`}. Sets
            how the `comp_edges` are built. If `source_and_sink`, an edge between
            compartments is kept only if source and sink compartments are within
            the view. If `source_or_sink`, an edge is kept if either the source
            or the sink are within the view. If `endpoint`, then the edge is kept
            if the compartment is in source or sink and if it is an edge between
            parent compartment and branchpoint. If `startpoint`, then the edge is
            kept if the compartment is in source or sink and if it is an edge
            between child compartment and branchpoint. This is used because we
            want different treatment of the `comp_edges` depending on whether we
            index with `.branch()` (`source_or_sink`), `.comp()`
            (`source_and_sink`), `.loc(0.0)` (`startpoint`), or `.loc(1.0)`
            (`endpoint`).

    Returns:
        A Sequence of indices of the `comp_edges` that are within the `View`.
    """
    pre = comp_edges["source"].isin(incl_comps).to_numpy()
    post = comp_edges["sink"].isin(incl_comps).to_numpy()
    comp_edge_inds = comp_edges.index.to_numpy()
    if comp_edge_condition == "source_and_sink":
        possible_edges_in_view = comp_edge_inds[(pre & post).flatten()]
    elif comp_edge_condition == "source_or_sink":
        possible_edges_in_view = comp_edge_inds[(pre | post).flatten()]
    elif comp_edge_condition == "startpoint":
        # Type 2 and 4 are comp_edges between branchpoints and the
        # child compartments.
        is_child = comp_edges["type"].isin([2, 4]).to_numpy()
        possible_edges_in_view = comp_edge_inds[((pre | post) & is_child).flatten()]
    elif comp_edge_condition == "endpoint":
        # Type 2 and 4 are comp_edges between branchpoints and the
        # parent compartments.
        is_parent = comp_edges["type"].isin([1, 3]).to_numpy()
        possible_edges_in_view = comp_edge_inds[((pre | post) & is_parent).flatten()]
    else:
        raise ValueError(
            f"comp_edge_condition is {comp_edge_condition}, but must be in "
            "{source_and_sink, source_or_sink, startpoint, endpoint}."
        )
    return possible_edges_in_view



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

from typing import List

import numpy as np
import pytest

import jaxley as jx
from jaxley.channels import HH
from jaxley.connect import connect
from jaxley.synapses import (
    IonotropicSynapse,
    Synapse,
    TanhConductanceSynapse,
    TanhRateSynapse,
    TestSynapse,
)


def test_multiparameter_setting(SimpleNet):
    """
    Test if the correct parameters are set if one type of synapses is inserted.

    Tests global index dropping: d4daaf019596589b9430219a15f1dda0b1c34d85
    """
    net = SimpleNet(2, 1, 4)

    pre = net.cell(0).branch(0).loc(0.0)
    post = net.cell(1).branch(0).loc(0.0)
    connect(pre, post, IonotropicSynapse())

    syn_view = net.IonotropicSynapse
    syn_params = ["IonotropicSynapse_gS", "IonotropicSynapse_e_syn"]

    for p in syn_params:
        syn_view.set(p, 0.32)


def _get_synapse_view(net, synapse_name, single_idx=1, double_idxs=[2, 3]):
    """Access to the synapse view"""
    if synapse_name == "IonotropicSynapse":
        full_syn_view = net.IonotropicSynapse
        single_syn_view = net.IonotropicSynapse.edge(single_idx)
        double_syn_view = net.IonotropicSynapse.edge(double_idxs)
    if synapse_name == "TanhRateSynapse":
        full_syn_view = net.TanhRateSynapse
        single_syn_view = net.TanhRateSynapse.edge(single_idx)
        double_syn_view = net.TanhRateSynapse.edge(double_idxs)
    if synapse_name == "TestSynapse":
        full_syn_view = net.TestSynapse
        single_syn_view = net.TestSynapse.edge(single_idx)
        double_syn_view = net.TestSynapse.edge(double_idxs)
    if synapse_name == "TanhConductanceSynapse":
        full_syn_view = net.TanhConductanceSynapse
        single_syn_view = net.TanhConductanceSynapse.edge(single_idx)
        double_syn_view = net.TanhConductanceSynapse.edge(double_idxs)
    return full_syn_view, single_syn_view, double_syn_view


@pytest.mark.parametrize(
    "synapse_type",
    [IonotropicSynapse, TanhRateSynapse, TestSynapse, TanhConductanceSynapse],
)
def test_set_and_querying_params_one_type(synapse_type, SimpleNet):
    """Test if the correct parameters are set if one type of synapses is inserted."""
    synapse_type = synapse_type()
    net = SimpleNet(4, 1, 4)

    for pre_ind in [0, 1]:
        for post_ind in [2, 3]:
            pre = net.cell(pre_ind).branch(0).loc(0.0)
            post = net.cell(post_ind).branch(0).loc(0.0)
            connect(pre, post, synapse_type)

    # Get the synapse parameters to test setting
    syn_params = list(synapse_type.synapse_params.keys())
    for p in syn_params:
        net.set(p, 0.15)
        assert np.all(net.edges[p].to_numpy() == 0.15)

    synapse_name = type(synapse_type).__name__
    full_syn_view, single_syn_view, double_syn_view = _get_synapse_view(
        net, synapse_name
    )

    # There shouldn't be too many synapse_params otherwise this will take a long time
    for p in syn_params:
        full_syn_view.set(p, 0.32)
        assert np.all(net.edges[p].to_numpy() == 0.32)

        single_syn_view.set(p, 0.18)
        assert net.edges[p].to_numpy()[1] == 0.18
        assert np.all(net.edges[p].to_numpy()[np.asarray([0, 2, 3])] == 0.32)

        double_syn_view.set(p, 0.12)
        assert net.edges[p][0] == 0.32
        assert net.edges[p][1] == 0.18
        assert np.all(net.edges[p].to_numpy()[np.asarray([2, 3])] == 0.12)


@pytest.mark.parametrize(
    "synapse_type", [TanhRateSynapse, TanhConductanceSynapse, TestSynapse]
)
def test_set_and_querying_params_two_types(synapse_type, SimpleNet):
    """Test whether the correct parameters are set."""
    synapse_type = synapse_type()
    net = SimpleNet(4, 1, 4)

    for pre_ind in [0, 1]:
        for post_ind, synapse in zip([2, 3], [IonotropicSynapse(), synapse_type]):
            pre = net.cell(pre_ind).branch(0).loc(0.0)
            post = net.cell(post_ind).branch(0).loc(0.0)
            connect(pre, post, synapse)

    type1_params = list(IonotropicSynapse().synapse_params.keys())
    synapse_type_params = list(synapse_type.synapse_params.keys())

    default_synapse_type = net.edges[synapse_type_params[0]].to_numpy()[[1, 3]]

    net.set(type1_params[0], 0.15)
    assert np.all(net.edges[type1_params[0]].to_numpy()[[0, 2]] == 0.15)
    if synapse_type_params[0] != type1_params[0]:
        assert np.all(
            net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == default_synapse_type
        )
    else:
        default_synapse_type = 0.15

    synapse_type_name = type(synapse_type).__name__
    synapse_type_full, synapse_type_single, synapse_type_double = _get_synapse_view(
        net, synapse_type_name, double_idxs=[0, 1]
    )

    # Generalize to all parameters
    net.IonotropicSynapse.set(type1_params[0], 0.32)
    assert np.all(net.edges[type1_params[0]].to_numpy()[[0, 2]] == 0.32)
    assert np.all(
        net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == default_synapse_type
    )

    synapse_type_full.set(synapse_type_params[0], 0.18)
    assert np.all(net.edges[type1_params[0]].to_numpy()[[0, 2]] == 0.32)
    assert np.all(net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == 0.18)

    net.IonotropicSynapse.edge(1).set(type1_params[0], 0.24)
    assert net.edges[type1_params[0]][0] == 0.32
    assert net.edges[type1_params[0]][2] == 0.24
    assert np.all(net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == 0.18)

    net.IonotropicSynapse.edge([0, 1]).set(type1_params[0], 0.27)
    assert np.all(net.edges[type1_params[0]].to_numpy()[[0, 2]] == 0.27)
    assert np.all(net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == 0.18)

    synapse_type_double.set(synapse_type_params[0], 0.21)
    assert np.all(net.edges[type1_params[0]].to_numpy()[[0, 2]] == 0.27)
    assert np.all(net.edges[synapse_type_params[0]].to_numpy()[[1, 3]] == 0.21)


@pytest.mark.parametrize(
    "synapse_type", [TanhRateSynapse, TanhConductanceSynapse, TestSynapse]
)
def test_shuffling_order_of_set(synapse_type, SimpleNet):
    """Test whether the result is the same if the order of synapses is changed."""
    synapse_type = synapse_type()

    net1 = SimpleNet(4, 1, 4)
    net2 = SimpleNet(4, 1, 4)

    connect(
        net1.cell(0).branch(0).loc(1.0),
        net1.cell(1).branch(0).loc(0.1),
        IonotropicSynapse(),
    )
    connect(
        net1.cell(1).branch(0).loc(0.6),
        net1.cell(2).branch(0).loc(0.7),
        IonotropicSynapse(),
    )
    connect(
        net1.cell(2).branch(0).loc(0.4), net1.cell(3).branch(0).loc(0.3), synapse_type
    )
    connect(
        net1.cell(3).branch(0).loc(0.1), net1.cell(1).branch(0).loc(0.1), synapse_type
    )

    # Different order as for `net1`.
    connect(
        net2.cell(3).branch(0).loc(0.1), net2.cell(1).branch(0).loc(0.1), synapse_type
    )
    connect(
        net2.cell(1).branch(0).loc(0.6),
        net2.cell(2).branch(0).loc(0.7),
        IonotropicSynapse(),
    )
    connect(
        net2.cell(2).branch(0).loc(0.4), net2.cell(3).branch(0).loc(0.3), synapse_type
    )
    connect(
        net2.cell(0).branch(0).loc(1.0),
        net2.cell(1).branch(0).loc(0.1),
        IonotropicSynapse(),
    )

    net1.insert(HH())
    net2.insert(HH())

    for i in range(4):
        net1.cell(i).branch(0).loc(0.0).record()
        net2.cell(i).branch(0).loc(0.0).record()

    voltages1 = jx.integrate(net1, t_max=5.0)
    voltages2 = jx.integrate(net2, t_max=5.0)

    assert np.max(np.abs(voltages1 - voltages2)) < 1e-8


def test_synapse_recording_order(SimpleNet):
    """Ensure that record finds the correct synapse index"""
    net1 = SimpleNet(4, 1, 1)
    jx.fully_connect(net1.cell([0, 1]), net1.cell([3]), IonotropicSynapse())
    net1.IonotropicSynapse.record("IonotropicSynapse_s")
    stim = np.zeros_like(np.arange(0, 2, 0.025))  # 2 ms stimulus
    stim[int(0.5 / 0.025) :] = 0.05
    net1.cell(1).stimulate(stim)
    soln1 = jx.integrate(net1)

    net2 = SimpleNet(4, 1, 1)
    # Addition of these extra synapses should not throw off the indices recorded
    jx.fully_connect(net2.cell([0, 1]), net2.cell([2]), TestSynapse())
    jx.fully_connect(net2.cell([0, 1]), net2.cell([3]), IonotropicSynapse())
    net2.IonotropicSynapse.record("IonotropicSynapse_s")
    net2.cell(1).stimulate(stim)
    soln2 = jx.integrate(net2)

    assert np.allclose(soln1, soln2)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import os

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax
import jax.numpy as jnp
import numpy as np
import pytest
from jax import jit

import jaxley as jx
from jaxley.channels import Leak
from jaxley.morphology import distance_direct, distance_pathwise


def test_direct_distance(SimpleCell):
    ncomp = 4
    length = 15.0

    cell = SimpleCell(5, ncomp)
    cell.branch("all").loc("all").set("length", length)
    cell.compute_xyz()
    cell.compute_compartment_centers()
    dist = distance_direct(cell.branch(0).loc(0.0), cell.branch(0).loc(1.0))
    assert dist[0] == (ncomp - 1) * length

    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=ncomp)
    cell = jx.Cell(branch, parents=[-1, 0, 1])
    cell.branch("all").loc("all").set("length", length)
    cell.compute_xyz()
    cell.compute_compartment_centers()
    dist = distance_direct(cell.branch(0).loc(0.0), cell.branch(2).loc(1.0))
    assert dist[0] == (3 * ncomp - 1) * length

    move_x = 220.0
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=ncomp)
    cell = jx.Cell(branch, parents=[-1, 0, 1])
    cell.branch("all").loc("all").set("length", length)
    net = jx.Network([cell for _ in range(2)])
    net.compute_xyz()
    net.cell(1).move(move_x, 0, 0)
    net.compute_compartment_centers()
    dist = distance_direct(
        net.cell(1).branch(0).loc(0.0), net.cell(0).branch(0).loc(0.0)
    )
    assert dist[0] == move_x

    dist = distance_direct(
        net.cell(0).branch(0).loc(0.0), net.cell(0).branch(0).loc(0.0)
    )
    assert dist[0] == 0.0

    dist = distance_direct(
        net.cell(1).branch(2).loc(0.3), net.cell(1).branch(2).loc(0.3)
    )
    assert dist[0] == 0.0

    dist = distance_direct(
        net.cell(1).branch(0).loc(0.0), net.cell(1).branch(2).loc(1.0)
    )
    assert dist[0] == (3 * ncomp - 1) * length


def test_pathwise_distance(SimpleCell):
    ncomp = 4
    length = 25.0
    cell = SimpleCell(5, ncomp)
    cell.branch("all").loc("all").set("length", length)
    pathdist = distance_pathwise(cell[0, 0], cell)
    cell.nodes["path_dist"] = pathdist

    dist_to_tip = cell[3, 3].nodes["path_dist"].item()
    assert (
        dist_to_tip == (ncomp * 3 - 1) * length
    ), f"{dist_to_tip} != {(ncomp * 3 - 1) * length}"


@pytest.mark.parametrize("kind", ["direct", "pathwise"])
def test_distance_within_jit(SimpleCell, kind: str):
    cell = SimpleCell(2, 3)
    cell.insert(Leak())
    cell.compute_xyz()
    cell.compute_compartment_centers()

    cell[0, 0].stimulate(0.1 * jnp.ones((100,)))
    cell[0, 0].record()

    if kind == "direct":
        dists = distance_direct(cell[0, 0], cell)
    elif kind == "pathwise":
        dists = distance_pathwise(cell[0, 0], cell)
    else:
        raise ValueError

    def simulate(sigmoid_offset, global_scaling):
        pstate = None
        counter = 0
        for branch in cell:
            for comp in branch:
                dist = dists[counter]
                conductance = global_scaling / (jnp.exp(-(dist + sigmoid_offset)))
                pstate = comp.data_set("Leak_gLeak", conductance * 1e-4, pstate)
        return jx.integrate(cell, param_state=pstate)

    sigmoid_offset = jnp.ones((1,))
    global_scaling = jnp.ones((1,))

    jitted_simulate = jit(simulate)
    v = jitted_simulate(sigmoid_offset, global_scaling)
    assert np.invert(np.any(np.isnan(v)))


def test_distance_swc(SimpleMorphCell):
    """Ensures that the distance computation for an SWC file remains unchanged."""
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", "morph_ca1_n120_250.swc")

    cell = SimpleMorphCell(
        fname,
        1,
        max_branch_len=2_000.0,
        ignore_swc_tracing_interruptions=True,
    )
    cell.set("axial_resistivity", 100.0)
    # Reasonable default values for most models.
    frequency = 100.0
    d_lambda = 0.1  # Larger -> more coarse-grained.

    for branch in cell.branches:
        diameter = 2 * branch.nodes["radius"].to_numpy()[0]
        c_m = branch.nodes["capacitance"].to_numpy()[0]
        r_a = branch.nodes["axial_resistivity"].to_numpy()[0]
        l = branch.nodes["length"].to_numpy()[0]

        lambda_f = 1e5 * np.sqrt(diameter / (4 * np.pi * frequency * c_m * r_a))
        ncomp = int((l / (d_lambda * lambda_f) + 0.9) / 2) * 2 + 1
        branch.set_ncomp(ncomp, initialize=False)

    # After the loop, you have to run `cell.initialize()` because we passed
    # `set_ncomp(..., initialize=False)` for speeding up the loop over branches.
    cell.initialize()

    dists_pathwise = distance_pathwise(cell.soma.branch(0).comp(0), cell)
    dists_direct_250610 = np.asarray(
        [
            0.0,
            368.06401336,
            692.67274231,
            864.42547931,
            806.69255969,
            743.73139166,
            760.54892662,
        ]
    )
    error = np.max(np.asarray(dists_pathwise)[::10] - dists_direct_250610)
    assert error < 1e-8, f"Error for pathwise distance is to large: {error} > 1e-8."

    dists_direct = distance_direct(cell.soma.branch(0).comp(0), cell)
    dists_direct_250610 = np.asarray(
        [
            0.0,
            194.3651027,
            383.63095437,
            467.20607951,
            483.88675182,
            413.21848932,
            428.11599931,
        ]
    )
    error = np.max(np.asarray(dists_direct)[::10] - dists_direct_250610)
    assert error < 1e-8, f"Error for direct distance is to large: {error} > 1e-8."



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import numpy as np


def is_same_network(pre: "View", post: "View") -> bool:
    """Check if views are from the same network."""
    is_in_net = "network" in pre.base.__class__.__name__.lower()
    is_in_same_net = pre.base is post.base
    return is_in_net and is_in_same_net


def sample_comp(cell_view: "View", num: int = 1, replace=True) -> "CompartmentView":
    """Sample a compartment from a cell.

    Returns View with shape (num, num_cols)."""
    return np.random.choice(cell_view._comps_in_view, num, replace=replace)


def connect(
    pre: "View",
    post: "View",
    synapse_type: "Synapse",
):
    """Connect specific compartments of a network with a synapse.

    The pre- and postsynaptic compartments must be compartments of the
    same network. If `pre` and `post` are both just a single compartment, then this
    function instantiates a single synapse. If `pre` and `post` are both `N`
    compartments, then this function instatiates `N` synapses (first-to-first,
    second-to-second,...).

    Args:
        pre: View of the presynaptic compartment.
        post: View of the postsynaptic compartment.
        synapse_type: The type of synapse to use.

    Example usage
    ^^^^^^^^^^^^^

    Example 1: Connect one compartment to another compartment with a single synapse:

    ::

        from jaxley.connect import connect
        from jaxley.synapses import IonotropicSynapse

        net = jx.Network([cell for _ in range(10)])
        connect(
            net.cell(0).branch(0).comp(0),
            net.cell(1).branch(0).comp(0),
            IonotropicSynapse(),
        )
        print(net.edges)

    Example 2: Connect `N` compartments to `N` other compartments with `N` synapses:

    ::

        from jaxley.connect import connect
        from jaxley.synapses import IonotropicSynapse

        net = jx.Network([cell for _ in range(10)])
        connect(
            net.cell(0).branch([0, 1]).comp(0),
            net.cell(1).branch([2, 3]).comp(0),
            IonotropicSynapse(),
        )
        print(net.edges)
    """
    assert is_same_network(
        pre, post
    ), "Pre and post compartments must be part of the same network."

    pre.base._append_multiple_synapses(pre.nodes, post.nodes, synapse_type)


def fully_connect(
    pre_cell_view: "View",
    post_cell_view: "View",
    synapse_type: "Synapse",
    random_post_comp: bool = False,
):
    """Fully (densely) connect cells of a network with synapses.

    Connections are from branch 0 location 0 of the pre-synaptic cell to branch 0
    location 0 of the post-synaptic cell unless random_post_comp=True.

    Args:
        pre_cell_view: View of the presynaptic cell.
        post_cell_view: View of the postsynaptic cell.
        synapse_type: The synapse to append.
        random_post_comp: If True, randomly samples the postsynaptic compartments.

    Example usage
    ^^^^^^^^^^^^^

    The following example insert 12 synapses (3 x 4).

    ::

        from jaxley.connect import fully_connect

        net = jx.Network([cell for _ in range(10)])
        fully_connect(
            net.cell([0, 1, 2]),
            net.cell([3, 4, 5, 6]),
            IonotropicSynapse(),
        )
        print(net.edges)
    """
    # Get pre- and postsynaptic cell indices.
    num_pre = len(pre_cell_view._cells_in_view)
    num_post = len(post_cell_view._cells_in_view)

    # Pre-synapse is at the zero-eth branch and zero-eth compartment.
    pre_rows = pre_cell_view.scope("local").branch(0).comp(0).nodes.copy()
    # Repeat rows `num_post` times. See SO 50788508.
    pre_rows = pre_rows.loc[pre_rows.index.repeat(num_post)]

    if random_post_comp:
        global_post_comp_indices = (
            post_cell_view.nodes.groupby("global_cell_index")
            .sample(num_pre, replace=True)
            .index.to_numpy()
        )
        # Reorder the post comp inds to tile order (pre indices are repeated so here tile needed)
        global_post_comp_indices = np.reshape(
            global_post_comp_indices, (num_pre, num_post)
        ).T.flatten()
    else:
        # Post-synapse also at the zero-eth branch and zero-eth compartment
        post_cell_view.nodes["orig_index"] = post_cell_view.nodes.index
        global_post_comp_indices = (
            post_cell_view.nodes.groupby("global_cell_index").first()["orig_index"]
        ).to_numpy()
        to_idx = np.tile(range(0, num_post), num_pre)
        global_post_comp_indices = global_post_comp_indices[to_idx]
        post_cell_view.nodes.drop(columns="orig_index", inplace=True)

    post_rows = post_cell_view.nodes.loc[global_post_comp_indices]

    pre_cell_view.base._append_multiple_synapses(pre_rows, post_rows, synapse_type)


def sparse_connect(
    pre_cell_view: "View",
    post_cell_view: "View",
    synapse_type: "Synapse",
    p: float,
    random_post_comp: bool = False,
):
    """Sparsely connect cells of a network with synapses.

    Connections are from branch 0 location 0 of the pre-synaptic cell to branch 0
    location 0 of the post-synaptic cell unless random_post_comp=True.

    NOTE: This function does not generate sparse random connectivity with random graph
    generation methodology. Cells may be connected multiple times and p=1.0 does
    not fully connect.

    Args:
        pre_cell_view: View of the presynaptic cell.
        post_cell_view: View of the postsynaptic cell.
        synapse_type: The synapse to append.
        p: Probability of connection.
        random_post_comp: If True, randomly samples the postsynaptic compartments.

    Example usage
    ^^^^^^^^^^^^^

    The following example insert approximately 6 synapses (3 x 4 = 12 possible
    synapses, with connection probability 0.5).

    ::

        from jaxley.connect import sparse_connect
        from jaxley.synapses import IonotropicSynapse

        net = jx.Network([cell for _ in range(10)])
        sparse_connect(
            net.cell([0, 1, 2]),
            net.cell([3, 4, 5, 6]),
            IonotropicSynapse(),
            p=0.5,
        )
        print(net.edges)
    """
    # Get pre- and postsynaptic cell indices.
    pre_cell_inds = pre_cell_view._cells_in_view
    post_cell_inds = post_cell_view._cells_in_view
    num_pre = len(pre_cell_view._cells_in_view)
    num_post = len(post_cell_view._cells_in_view)

    num_connections = np.random.binomial(num_pre * num_post, p)

    if num_connections == 0:
        # Don't do any of the following connections if no synapse is inserted.
        return
    pre_syn_neurons = np.random.choice(pre_cell_inds, size=num_connections)
    post_syn_neurons = np.random.choice(post_cell_inds, size=num_connections)

    # Sort the synapses only for convenience of inspecting `.edges`.
    sorting = np.argsort(pre_syn_neurons)
    pre_syn_neurons = pre_syn_neurons[sorting]
    post_syn_neurons = post_syn_neurons[sorting]

    pre_syn_neurons, inverse_pre = np.unique(pre_syn_neurons, return_inverse=True)

    # Pre-synapse is at the zero-eth branch and zero-eth compartment.
    global_pre_indices = (
        pre_cell_view.scope("global")
        .cell(pre_syn_neurons)
        .scope("local")
        .branch(0)
        .comp(0)
        .nodes.index
    )
    global_pre_indices = global_pre_indices[inverse_pre]
    pre_rows = pre_cell_view.base.nodes.loc[global_pre_indices]

    # Sample the post-synaptic compartments
    if random_post_comp:
        # Filter the post cell view to include post-synaptic neurons
        post_syn_view = post_cell_view.nodes[
            post_cell_view.nodes["global_cell_index"].isin(post_syn_neurons)
        ].copy()
        # Determine how many comps to sample for each post-synaptic neuron
        unique_cells, counts = np.unique(post_syn_neurons, return_counts=True)
        n_samples_dict = dict(zip(unique_cells, counts))
        post_syn_view["orig_index"] = post_syn_view.index
        sampled_inds = post_syn_view.groupby("global_cell_index").apply(
            lambda x: x.sample(n=n_samples_dict[x.name], replace=True)
        )
        global_post_comp_indices = sampled_inds.orig_index.to_numpy()
        post_rows = post_cell_view.nodes.loc[global_post_comp_indices]
        post_syn_view.drop(columns="orig_index", inplace=True)
    else:
        post_syn_neurons, inverse_post = np.unique(
            post_syn_neurons, return_inverse=True
        )
        # Post-synapse also at the zero-eth branch and zero-eth compartment
        global_post_indices = (
            post_cell_view.scope("global")
            .cell(post_syn_neurons)
            .scope("local")
            .branch(0)
            .comp(0)
            .nodes.index
        )
        global_post_indices = global_post_indices[inverse_post]
        post_rows = post_cell_view.base.nodes.loc[global_post_indices]

    if len(pre_rows) > 0:
        pre_cell_view.base._append_multiple_synapses(pre_rows, post_rows, synapse_type)


def connectivity_matrix_connect(
    pre_cell_view: "View",
    post_cell_view: "View",
    synapse_type: "Synapse",
    connectivity_matrix: np.ndarray[bool],
    random_post_comp: bool = False,
):
    """Connect cells of a network with synapses via a boolean connectivity matrix.

    Entries > 0 in the matrix indicate a connection between the corresponding cells.
    Connections are from branch 0 location 0 of the pre-synaptic cell to branch 0
    location 0 of the post-synaptic cell unless random_post_comp=True.

    Args:
        pre_cell_view: View of the presynaptic cell.
        post_cell_view: View of the postsynaptic cell.
        synapse_type: The synapse to append.
        connectivity_matrix: A boolean matrix indicating the connections between cells.
            If floating point values are passed, they are _not_ interpreted as
            synaptic weights, but we only check if they are zero (no connection) or
            not (connection).
        random_post_comp: If True, randomly samples the postsynaptic compartments.

    Example usage
    ^^^^^^^^^^^^^

    The following generates a random 10 x 10 boolean matrix and uses it to connect the
    neurons in a network.

    ::

        from jaxley.connect import connectivity_matrix_connect
        from jaxley.synapses import IonotropicSynapse

        net = jx.Network([cell for _ in range(10)])
        connectivity_matrix = np.random.choice([False, True], size=(10, 10))
        connectivity_matrix_connect(
            net.cell("all"),
            net.cell("all"),
            IonotropicSynapse(),
            connectivity_matrix,
        )
        print(net.edges)
    """
    # Get pre- and postsynaptic cell indices
    num_pre = len(pre_cell_view._cells_in_view)
    num_post = len(post_cell_view._cells_in_view)

    assert connectivity_matrix.shape == (
        num_pre,
        num_post,
    ), "Connectivity matrix must have shape (num_pre, num_post)."
    assert connectivity_matrix.dtype == bool, "Connectivity matrix must be boolean."

    # Get pre to post connection pairs from connectivity matrix
    from_idx, to_idx = np.where(connectivity_matrix)

    # Pre-synapse at the zero-eth branch and zero-eth compartment
    pre_cell_view.nodes["orig_index"] = pre_cell_view.nodes.index
    post_cell_view.nodes["orig_index"] = post_cell_view.nodes.index
    global_pre_comp_indices = (
        pre_cell_view.nodes.groupby("global_cell_index").first()["orig_index"]
    ).to_numpy()
    pre_rows = pre_cell_view.select(nodes=global_pre_comp_indices[from_idx]).nodes
    pre_cell_view.nodes.drop(columns="orig_index", inplace=True)

    if random_post_comp:
        global_to_idx = post_cell_view.nodes.global_cell_index.unique()[to_idx]
        # Filter the post cell view to include post-synaptic neurons selected
        post_syn_view = post_cell_view.nodes[
            post_cell_view.nodes["global_cell_index"].isin(global_to_idx)
        ]
        # Determine how many comps to sample for each post-synaptic neuron
        unique_cells, counts = np.unique(global_to_idx, return_counts=True)
        # Sample the post-synaptic compartments
        n_samples_dict = dict(zip(unique_cells, counts))
        sampled_inds = post_syn_view.groupby("global_cell_index").apply(
            lambda x: x.sample(n=n_samples_dict[x.name], replace=True)
        )
        global_post_comp_indices = sampled_inds.orig_index.to_numpy()
    else:
        # Post-synapse also at the zero-eth branch and zero-eth compartment
        global_post_comp_indices = (
            post_cell_view.nodes.groupby("global_cell_index").first()["orig_index"]
        ).to_numpy()
        global_post_comp_indices = global_post_comp_indices[to_idx]
        post_cell_view.nodes.drop(columns="orig_index", inplace=True)
    post_rows = post_cell_view.select(nodes=global_post_comp_indices).nodes

    pre_cell_view.base._append_multiple_synapses(pre_rows, post_rows, synapse_type)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Dict, Optional, Tuple

import jax.numpy as jnp

from jaxley.solver_gate import save_exp
from jaxley.synapses.synapse import Synapse


class TestSynapse(Synapse):
    """
    Compute syanptic current and update synapse state for a test synapse.
    """

    __test__ = False  # Not a unit test - pytest ignores

    def __init__(self, name: Optional[str] = None):
        super().__init__(name)
        prefix = self._name
        self.synapse_params = {f"{prefix}_gC": 1e-4}
        self.synapse_states = {f"{prefix}_c": 0.2}

    def update_states(
        self,
        states: Dict,
        delta_t: float,
        pre_voltage: float,
        post_voltage: float,
        params: Dict,
    ) -> Dict:
        """Return updated synapse state and current."""
        prefix = self._name
        v_th = -35.0
        delta = 10.0
        k_minus = 1.0 / 40.0

        s_bar = 1.0 / (1.0 + save_exp((v_th - pre_voltage) / delta))
        tau_s = (1.0 - s_bar) / k_minus

        s_inf = s_bar
        slope = -1.0 / tau_s
        exp_term = save_exp(slope * delta_t)
        new_s = states[f"{prefix}_c"] * exp_term + s_inf * (1.0 - exp_term)
        return {f"{prefix}_c": new_s}

    def compute_current(
        self, states: Dict, pre_voltage: float, post_voltage: float, params: Dict
    ) -> float:
        prefix = self._name
        e_syn = 0.0
        g_syn = params[f"{prefix}_gC"] * states[f"{prefix}_c"]
        return g_syn * (post_voltage - e_syn)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>
from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple

from jax import Array
from jax.typing import ArrayLike


class Pump:
    """Pump base class. All pumps inherit from this class.

    A pump in Jaxley is everything that modifies the intracellular ion concentrations.
    """

    _name: str
    channel_params: dict[str, Array]
    channel_states: dict[str, Array]
    current_name: str

    def __init__(self, name: Optional[str] = None):
        self._name = name if name else self.__class__.__name__

    @property
    def name(self) -> Optional[str]:
        """The name of the channel (by default, this is the class name)."""
        return self._name

    def change_name(self, new_name: str):
        """Change the pump name.

        Args:
            new_name: The new name of the pump.

        Returns:
            Renamed pump, such that this function is chainable.
        """
        old_prefix = self._name + "_"
        new_prefix = new_name + "_"

        self._name = new_name
        self.channel_params = {
            (
                new_prefix + key[len(old_prefix) :]
                if key.startswith(old_prefix)
                else key
            ): value
            for key, value in self.channel_params.items()
        }

        self.channel_states = {
            (
                new_prefix + key[len(old_prefix) :]
                if key.startswith(old_prefix)
                else key
            ): value
            for key, value in self.channel_states.items()
        }
        return self

    def update_states(
        self, states: dict[str, Array], v, params: dict[str, Array]
    ) -> None:
        """Update the states of the pump."""
        raise NotImplementedError

    def compute_current(
        self, states: dict[str, Array], v, params: dict[str, Array]
    ) -> None:
        """Given channel states and voltage, return the change in ion concentration.

        Args:
            states: All states of the compartment.
            v: Voltage of the compartment in mV.
            params: Parameters of the channel (conductances in `S/cm2`).

        Returns:
            Ion concentration change in `mM`.
        """
        raise NotImplementedError

    def init_state(
        self,
        states: dict[str, ArrayLike],
        v: ArrayLike,
        params: dict[str, ArrayLike],
        delta_t: float,
    ):
        """Initialize states of channel."""
        return {}



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>
from typing import Optional

import jax
import pytest
from jax import Array
from jax.typing import ArrayLike

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax.numpy as jnp
import numpy as np
from jaxley_mech.channels.l5pc import CaHVA, CaLVA
from jaxley_mech.channels.l5pc import CaPump as CaPumpAsChannel

import jaxley as jx
from jaxley.channels import HH, K, Leak, Na
from jaxley.channels.channel import Channel
from jaxley.pumps import CaFaradayConcentrationChange, CaNernstReversal, CaPump, Pump


class NaPump(Pump):
    """A dummy pump which is only used for internal `Jaxley` tests.

    Modeled after the calcium channel in Destexhe et al. 1994.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name)
        self.channel_params = {
            f"{self._name}_gamma": 0.05,  # Fraction of free calcium (not buffered).
            f"{self._name}_decay": 80,  # Buffering time constant in ms.
            f"{self._name}_depth": 0.1,  # Depth of shell in um.
            f"{self._name}_minNai": 1e-4,  # Minimum intracell. ca concentration in mM.
        }
        self.channel_states = {"i_Na": 1e-8, "NaCon_i": 5e-05}
        self.ion_name = "NaCon_i"
        self.current_name = "i_NaPump"
        self.META = {
            "reference": "Made up by Jaxley developers (@michaeldeistler)",
            "mechanism": "Sodium dynamics",
        }

    def update_states(
        self,
        states: dict[str, Array],
        dt,
        v,
        params: dict[str, Array],
    ):
        """Update states if necessary (but this pump has no states to update)."""
        return {"NaCon_i": states["NaCon_i"], "i_Na": states["i_Na"]}

    def compute_current(
        self,
        states: dict[str, Array],
        modified_state,
        params: dict[str, Array],
    ):
        """Return change of calcium concentration based on calcium current and decay."""
        prefix = self._name
        ica = states["i_Na"] / 1_000.0
        gamma = params[f"{prefix}_gamma"]
        decay = params[f"{prefix}_decay"]
        depth = params[f"{prefix}_depth"]
        minCai = params[f"{prefix}_minNai"]

        FARADAY = 96485  # Coulombs per mole.

        # Calculate the contribution of calcium currents to cai change.
        drive_channel = -10_000.0 * ica * gamma / (2 * FARADAY * depth)
        state_decay = (modified_state - minCai) / decay
        diff = drive_channel - state_decay
        return -diff

    def init_state(
        self,
        states: dict[str, ArrayLike],
        v: ArrayLike,
        params: dict[str, ArrayLike],
        delta_t: float,
    ):
        """Initialize states of channel."""
        return {}


class NaNernstReversal(Channel):
    """Compute Calcium reversal from inner and outer concentration of calcium."""

    def __init__(
        self,
        name: Optional[str] = None,
    ):
        self.current_is_in_mA_per_cm2 = True
        super().__init__(name)
        self.channel_constants = {
            "F": 96485.3329,  # C/mol (Faraday's constant)
            "T": 279.45,  # Kelvin (temperature)
            "R": 8.314,  # J/(mol K) (gas constant)
        }
        self.channel_params = {}
        self.channel_states = {"eNa": 0.0, "NaCon_i": 5e-05, "NaCon_e": 3.0}
        self.current_name = f"i_Na"
        self.META = {"ion": "Na"}

    def update_states(self, u, dt, voltages, params):
        """Update internal calcium concentration based on calcium current and decay."""
        R, T, F = (
            self.channel_constants["R"],
            self.channel_constants["T"],
            self.channel_constants["F"],
        )
        Nai = u["NaCon_i"]
        Nao = u["NaCon_e"]
        C = R * T / (2 * F) * 1000  # mV
        vNa = C * jnp.log(Nao / Nai)
        return {"eNa": vNa, "NaCon_i": Nai, "NaCon_e": Nao}

    def compute_current(self, u, voltages, params):
        """This dynamics model does not directly contribute to the membrane current."""
        return 0

    def init_state(self, states, voltages, params, delta_t):
        """Initialize the state at fixed point of gate dynamics."""
        return {}


class CaPump2(Pump):
    """A copy of the Calcium dynamics tracking inside calcium concentration.

    Modeled after Destexhe et al. 1994.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name)
        self.channel_params = {
            f"{self._name}_gamma": 0.05,  # Fraction of free calcium (not buffered).
            f"{self._name}_decay": 80,  # Buffering time constant in ms.
            f"{self._name}_depth": 0.1,  # Depth of shell in um.
            f"{self._name}_minCai": 1e-4,  # Minimum intracell. ca concentration in mM.
        }
        self.channel_states = {"i_Ca": 1e-8, "CaCon_i": 5e-05}
        self.ion_name = "CaCon_i"
        self.current_name = "i_CaPump"
        self.META = {
            "reference": "Modified from Destexhe et al., 1994",
            "mechanism": "Calcium dynamics",
        }

    def update_states(
        self,
        states: dict[str, Array],
        dt,
        v,
        params: dict[str, Array],
    ):
        """Update states if necessary (but this pump has no states to update)."""
        return {"CaCon_i": states["CaCon_i"], "i_Ca": states["i_Ca"]}

    def compute_current(
        self,
        states: dict[str, Array],
        modified_state,
        params: dict[str, Array],
    ):
        """Return change of calcium concentration based on calcium current and decay."""
        prefix = self._name
        ica = states["i_Ca"] / 1_000.0
        gamma = params[f"{prefix}_gamma"]
        decay = params[f"{prefix}_decay"]
        depth = params[f"{prefix}_depth"]
        minCai = params[f"{prefix}_minCai"]

        FARADAY = 96485  # Coulombs per mole.

        # Calculate the contribution of calcium currents to cai change.
        drive_channel = -10_000.0 * ica * gamma / (2 * FARADAY * depth)
        state_decay = (modified_state - minCai) / decay
        diff = drive_channel - state_decay
        return -diff

    def init_state(
        self,
        states: dict[str, ArrayLike],
        v: ArrayLike,
        params: dict[str, ArrayLike],
        delta_t: float,
    ):
        """Initialize states of channel."""
        return {}


def _build_active_cell():
    """Helper that builds an active cell with Na, K, Leak separately."""
    current = jx.step_current(0.5, 3.0, 0.1, 0.025, 10.0)
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=2)
    cell = jx.Cell(branch, parents=[-1, 0])
    cell.branch(0).comp(0).stimulate(current)

    # Insert channels.
    cell.insert(Na())
    cell.insert(K())
    cell.insert(Leak())
    cell.insert(CaHVA())
    cell.insert(NaNernstReversal())
    cell.insert(CaNernstReversal())

    # Recordings and simulation.
    cell.branch(1).comp(1).record("v")
    cell.branch(0).comp(1).record("CaCon_i")
    cell.branch(0).comp(1).record("NaCon_i")
    return cell


def test_that_order_of_insert_does_not_matter():
    """Insert pumps in a different orders and check if all give the same simulation."""
    ############################ Option 1:
    cell = _build_active_cell()
    cell.insert(CaPump())
    # Also test `delete` for a pump.
    cell.insert(CaFaradayConcentrationChange())
    cell.delete(CaFaradayConcentrationChange())
    # Return to the channel that should really be inserted.
    cell.insert(NaPump())
    cell.insert(CaPump2())
    cell.diffuse("CaCon_i")
    cell.diffuse("NaCon_i")
    cell.set("CaHVA_gCaHVA", 0.0001)
    cell.set("axial_diffusion_CaCon_i", 200.0)
    cell.set("axial_diffusion_NaCon_i", 400.0)
    cell.init_states()
    v1 = jx.integrate(cell, voltage_solver="jaxley.dhs.cpu")

    ############################ Option 2:
    cell = _build_active_cell()
    cell.insert(CaPump())
    cell.diffuse("CaCon_i")
    cell.insert(NaPump())
    cell.insert(CaPump2())
    cell.diffuse("NaCon_i")
    # Delete and diffuse again to test `delete_diffusion` method.
    cell.delete_diffusion("NaCon_i")
    cell.diffuse("NaCon_i")
    cell.set("CaHVA_gCaHVA", 0.0001)
    cell.set("axial_diffusion_CaCon_i", 200.0)
    cell.set("axial_diffusion_NaCon_i", 400.0)
    cell.init_states()
    v2 = jx.integrate(cell, voltage_solver="jaxley.dhs.cpu")

    ############################ Option 3:
    cell = _build_active_cell()
    cell.insert(CaPump())
    cell.insert(NaPump())
    cell.diffuse("NaCon_i")
    cell.insert(CaPump2())
    cell.diffuse("CaCon_i")
    cell.set("CaHVA_gCaHVA", 0.0001)
    cell.set("axial_diffusion_CaCon_i", 200.0)
    cell.set("axial_diffusion_NaCon_i", 400.0)
    cell.init_states()
    v3 = jx.integrate(cell, voltage_solver="jaxley.dhs.cpu")

    ############################ Option 4:
    cell = _build_active_cell()
    cell.insert(CaPump2())
    cell.insert(CaPump())
    cell.insert(NaPump())
    cell.diffuse("CaCon_i")
    cell.diffuse("NaCon_i")
    cell.set("CaHVA_gCaHVA", 0.0001)
    cell.set("axial_diffusion_CaCon_i", 200.0)
    cell.set("axial_diffusion_NaCon_i", 400.0)
    cell.init_states()
    v4 = jx.integrate(cell, voltage_solver="jaxley.dhs.cpu")

    for i, v_compare in enumerate([v2, v3, v4]):
        max_error = np.max(np.abs(v1 - v_compare))
        assert max_error < 1e-8, f"Gap between v1 and v{i+1} is {max_error}."


def _build_calcium_cell():
    """Helper that builds a cell with relevant channels."""
    current = jx.step_current(0.5, 3.0, 0.1, 0.025, 5.0)
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=2)
    cell = jx.Cell(branch, parents=[-1, 0])
    cell.branch(0).comp(0).stimulate(current)

    # Insert channels.
    cell.insert(HH())
    cell.insert(CaHVA())
    cell.insert(CaNernstReversal())

    # Recordings and simulation.
    cell.branch(0).comp(1).record("CaCon_i")
    return cell


def test_that_high_axial_diffusion_approaches_no_diffusion():
    """Test if very high axial ion resisitvity approaches no diffusion at all.

    By the side, this test also implicitly tests the `cranck_nicolson` solver in
    combination with `jax.sparse`.
    """
    cell1 = _build_calcium_cell()
    cell1.insert(CaPump())
    cell1.diffuse("CaCon_i")
    # Very high axial resistivity, almost no diffusion.
    cell1.set("axial_diffusion_CaCon_i", 1e-8)
    v1 = jx.integrate(cell1, solver="crank_nicolson", voltage_solver="jaxley.dhs.cpu")

    cell2 = _build_calcium_cell()
    cell2.insert(CaPump())
    v2 = jx.integrate(cell2, solver="crank_nicolson", voltage_solver="jaxley.dhs.cpu")

    max_error = np.max(np.abs(v1 - v2))
    assert max_error < 1e-7, f"Gap is {max_error}."


def test_pump_exists_only_in_parts_of_the_cell():
    """Test that there is no NaN when a pump exists only in parts of a cell.

    Diffusion can still introduce the new molecule to the rest of the cell.

    By the side, this test also implicitly tests the `cranck_nicolson` solver in
    combination with `jaxley.stone` (default).
    """
    current = jx.step_current(0.5, 3.0, 0.1, 0.025, 5.0)
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=4)
    cell = jx.Cell(branch, parents=[-1, 0])
    cell.branch(0).comp(0).stimulate(current)

    # Insert channels.
    cell.insert(HH())

    # Calcium only in branch(1)
    cell.branch(1).insert(CaHVA())
    cell.branch(1).insert(CaNernstReversal())
    cell.branch(1).insert(CaPump())
    cell.diffuse("CaCon_i")

    # Very high axial resistivity, almost no diffusion.
    cell.record("CaCon_i")
    recordings = jx.integrate(cell, solver="crank_nicolson")
    assert np.invert(np.any(np.isnan(recordings))), "Found NaN value!"


def test_diffusion_in_parts_of_cell_raises():
    """Assert the raise if one runs diffusion on parts of a cell."""
    cell = _build_calcium_cell()

    with pytest.raises(AssertionError):
        cell.branch(0).diffuse("CaCon_i")


def test_raise_ion_diffusion_in_one_cell_of_net1():
    """Assert the raise if a network is built of cells that are diffused."""
    cell1 = _build_calcium_cell()
    cell1.diffuse("CaCon_i")
    cell2 = _build_calcium_cell()
    with pytest.raises(AssertionError):
        net = jx.Network([cell1, cell2])


def test_raise_ion_diffusion_in_one_cell_of_net2():
    """Assert the raise if one runs diffusion on parts of a network."""
    cell1 = _build_calcium_cell()
    cell2 = _build_calcium_cell()
    net = jx.Network([cell1, cell2])
    with pytest.raises(AssertionError):
        net.cell(0).diffuse("CaCon_i")


def test_ion_is_diffused_but_not_pumped():
    """Test if, when an ion is diffused but not pumped, comps converge to one value."""
    cell = _build_calcium_cell()
    cell.diffuse("CaCon_i")
    # Initialize the calcium concentration to be different at the beginning.
    cell.branch(0).set("CaCon_i", 0.2)
    cell.branch(1).set("CaCon_i", 0.1)
    cell.set("axial_diffusion_CaCon_i", 1000.0)  # Relatively high diffusion.
    cell.record("CaCon_i")  # Record everywhere.

    # Run simulation and assert that calcium becomes the same everywhere.
    cacon_i = jx.integrate(cell, t_max=5.0)
    last_timepoint_cacon_i = cacon_i[:, -1]
    avg_cacon_i = np.mean(last_timepoint_cacon_i)
    max_diff_to_mean = np.max(np.abs(avg_cacon_i - last_timepoint_cacon_i))
    assert max_diff_to_mean < 1e-4, f"Error {max_diff_to_mean} > 1e-4"


def test_ion_is_pumped_but_not_diffused():
    """Test whether there is no NaN when an ion is pumped but not diffused."""
    cell = _build_calcium_cell()
    cacon_i = jx.integrate(cell, t_max=5.0, voltage_solver="jaxley.dhs.cpu")
    assert np.invert(np.any(np.isnan(cacon_i))), "Found NaN when ion is not diffused!"


def test_ion_diffusion_compartment():
    """Test whether `diffuse()` on a compartment does not lead to NaN."""
    comp = jx.Compartment()
    comp.insert(CaHVA())
    comp.insert(CaNernstReversal())
    comp.diffuse("CaCon_i")
    comp.set("axial_diffusion_CaCon_i", 1.0)
    comp.record("v")
    comp.record("CaCon_i")
    recs = jx.integrate(comp, t_max=10.0, voltage_solver="jaxley.dhs.cpu")
    assert np.invert(np.any(np.isnan(recs)))


def test_ion_diffusion_branch():
    """Test whether `diffuse()` on a branch does not lead to NaN.

    By the side, this tests the `fwd_euler` solver together with ion diffusion.
    """
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=2)
    branch.insert(CaHVA())
    branch.insert(CaNernstReversal())
    branch.diffuse("CaCon_i")
    branch.set("axial_diffusion_CaCon_i", 1.0)
    branch.record("v")
    branch.record("CaCon_i")
    recs = jx.integrate(branch, t_max=10.0, solver="fwd_euler")
    assert np.invert(np.any(np.isnan(recs)))


def test_ion_diffusion_cell():
    """Test whether `diffuse()` on a cell does not lead to NaN."""
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=2)
    cell = jx.Cell(branch, parents=[-1, 0, 0])
    cell.insert(CaHVA())
    cell.insert(CaNernstReversal())
    cell.diffuse("CaCon_i")
    cell.set("axial_diffusion_CaCon_i", 1.0)
    cell.record("v")
    cell.record("CaCon_i")
    recs = jx.integrate(cell, t_max=5.0, voltage_solver="jaxley.dhs.cpu")
    assert np.invert(np.any(np.isnan(recs)))


def test_ion_diffusion_net():
    """Test whether `diffuse()` on a network does not lead to NaN."""
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=2)
    cell = jx.Cell(branch, parents=[-1, 0])
    net = jx.Network([cell, cell])
    net.insert(CaHVA())
    net.insert(CaNernstReversal())
    net.diffuse("CaCon_i")
    net.set("axial_diffusion_CaCon_i", 1.0)
    net.record("v")
    net.record("CaCon_i")
    recs = jx.integrate(net, t_max=5.0, voltage_solver="jaxley.dhs.cpu")
    assert np.invert(np.any(np.isnan(recs)))


def test_diffuse_a_channel_state():
    """Test whether diffusing a channel state gives almost same results as pump.

    For full numerical accuracy, channel states should not be diffused (but only things)
    that are pumped). However, we still allow to diffuse channel states. This test
    evaluates whether the error of doing this is small enough.
    """
    cell1 = _build_calcium_cell()
    cell2 = _build_calcium_cell()
    cell1.insert(CaPump())
    cell2.insert(CaPumpAsChannel())

    for cell in [cell1, cell2]:
        cell.set("CaCon_i", 0.0001)
        cell.diffuse("CaCon_i")
        cell.set("axial_diffusion_CaCon_i", 1.0)
        cell.record("v")
        cell.record("CaCon_i")
    v1 = jx.integrate(cell1, t_max=10.0, voltage_solver="jaxley.dhs.cpu")
    v2 = jx.integrate(cell2, t_max=10.0, voltage_solver="jaxley.dhs.cpu")
    max_error = np.max(np.abs(v1 - v2))
    assert max_error < 1e-2, f"Error {max_error} > 1e-2"



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>
from typing import Any, List, Optional, Tuple

import jax.numpy as jnp
import networkx as nx
import numpy as np
import pandas as pd
from jax import Array


def convert_to_csc(
    num_elements: int, row_ind: np.ndarray, col_ind: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Convert between two representations for sparse systems.

    This is needed because `jax.scipy.linalg.spsolve` requires the `(ind, indptr)`
    representation, but the `(row, col)` is more intuitive and easier to build.

    This function uses `np` instead of `jnp` because it only deals with indexing which
    can be dealt with only based on the branch structure (i.e. independent of any
    parameter values).

    Written by ChatGPT.
    """
    data_inds = np.arange(num_elements)
    # Step 1: Sort by (col_ind, row_ind)
    sorted_indices = np.lexsort((row_ind, col_ind))
    data_inds = data_inds[sorted_indices]
    row_ind = row_ind[sorted_indices]
    col_ind = col_ind[sorted_indices]

    # Step 2: Create indptr array
    n_cols = col_ind.max() + 1
    indptr = np.zeros(n_cols + 1, dtype=int)
    np.add.at(indptr, col_ind + 1, 1)
    np.cumsum(indptr, out=indptr)

    # Step 3: The row indices are already sorted
    indices = row_ind

    return data_inds, indices, indptr


def comp_edges_to_indices(
    comp_edges: pd.DataFrame,
) -> tuple[int, Array, Array, Array]:
    """Generates sparse matrix indices from the table of node edges.

    This is only used for the `jax.sparse` voltage solver.

    Args:
        comp_edges: Dataframe with three columns (sink, source, type).

    Returns:
        n_nodes: The number of total nodes (including branchpoints).
        data_inds: The indices to reorder the data.
        indices and indptr: Indices passed to the sparse matrix solver.
    """
    # Build indices for diagonals.
    sources = np.asarray(comp_edges["source"].to_list())
    sinks = np.asarray(comp_edges["sink"].to_list())
    n_nodes = np.max(sinks) + 1 if len(sinks) > 0 else 1
    diagonal_inds = jnp.stack([jnp.arange(n_nodes), jnp.arange(n_nodes)])

    # Build indices for off-diagonals.
    off_diagonal_inds = jnp.stack([sources, sinks]).astype(int)

    # Concatenate indices of diagonals and off-diagonals.
    all_inds = jnp.concatenate([diagonal_inds, off_diagonal_inds], axis=1)

    # Cast (row, col) indices to the format required for the `jax` sparse solver.
    data_inds, indices, indptr = convert_to_csc(
        num_elements=all_inds.shape[1],
        row_ind=all_inds[0],
        col_ind=all_inds[1],
    )
    return data_inds, indices, indptr


def dhs_permutation_indices(
    lowers_and_uppers,
    offdiag_inds,
    node_order,
    mapping_dict,
):
    """Return mapping to the DHS solve order, also for lower and upper diagonal vals."""
    edge_data = {}
    for edge, v in zip(offdiag_inds.T, lowers_and_uppers):
        edge_data[tuple(edge.tolist())] = v

    ordered_comp_edges = {}
    for i, j in edge_data.keys():
        new_i = mapping_dict[i]
        new_j = mapping_dict[j]
        ordered_comp_edges[(new_i, new_j)] = edge_data[(i, j)]

    lowers = {}
    uppers = {}
    for ij in ordered_comp_edges.keys():
        if ij[0] < ij[1]:
            lowers[ij] = ordered_comp_edges[ij]
        else:
            uppers[ij] = ordered_comp_edges[ij]

    # We have to sort lowers and uppers to be in the solve order.
    children = []
    for key in lowers.keys():
        children.append(key[1])
    lower_sorting = np.argsort(children)

    parents = []
    for key in uppers.keys():
        parents.append(key[0])
    upper_sorting = np.argsort(parents)

    lowers = jnp.asarray(list(lowers.values()))[lower_sorting]
    uppers = jnp.asarray(list(uppers.values()))[upper_sorting]

    # Adapt node order.
    new_node_order = []
    for n in node_order:
        new_node_order.append(
            [mapping_dict[int(n[0])], mapping_dict[int(n[1])], int(n[2])]
        )
    # We need `.reshape(-1, 3)` because `new_node_order` is empty for point-neurons.
    # However, for networks, we still want to be able to concatenate this empty array
    # with `(N, 3)` arrays (e.g., from morph detailed neurons). The line below ensures
    # that `new_node_order` is of shape (0, 3) if it is empty.
    new_node_order = jnp.asarray(new_node_order, dtype=int).reshape(-1, 3)

    lowers_and_uppers = jnp.concatenate([lowers, uppers])
    return lowers_and_uppers, new_node_order


def dhs_solve_index(
    solve_graph: nx.DiGraph,
    allowed_nodes_per_level: int = 1,
    root: Optional[int] = 0,
) -> nx.DiGraph:
    """Given a compartment graph, return a directed graph indicating the solve order.

    Args:
        comp_graph: Compartment graph. Must have the compartment indices such that
            they match the ones in the `jx.Module` (i.e., the compartment graph must
            have been generated with `to_graph()` or it must have run
            `_set_comp_and_branch_index()` after having run
            `build_compartment_graph()`).
        root: The root node to traverse the graph for the DHS solve order.
        allowed_nodes_per_level: How many nodes are visited before the level is
            increased, even if the number of hops did not change.

    Returns:
        - node_and_parent: A list of tuples (node, parent, level), where the node and
        parent indicate the (child and parent) node index, and the level indicates the
        number of hops it took from the root to get there. The number of hops is
        currently unused, but it will be important in the future to parallelization.
        - node_to_solve_index_mapping: An dictionary mapping from the compartment
        indices to the solve indices.
    """
    undirected_solve_graph = solve_graph.to_undirected()

    # Traverse the graph for the solve order.
    # `sort_neighbors=lambda x: sorted(x)` to first handle nodes with lower node index.
    node_and_parent = [(root, -1, 0)]
    solve_graph.nodes[root]["solve_index"] = 0
    solve_index = 1
    for i, j, level in bfs_edge_hops(
        undirected_solve_graph, root, allowed_nodes_per_level
    ):
        solve_graph.add_edge(i, j)
        # Copy comp_index and branch_index from compartment graph. We only update the
        # solve_index.
        solve_graph.nodes[j]["solve_index"] = solve_index
        node_and_parent.append((j, i, level))
        solve_index += 1

    # Create a dictionary which maps every node to its solve index. Two notes:
    # - The node name corresponds to the compartment index (and branchpoints continue
    # numerically from where the compartments have ended)
    # - The solve index specifies the order in which the node is processed during the
    # DHS solve.
    inds = {node: solve_graph.nodes[node]["solve_index"] for node in solve_graph.nodes}
    node_to_solve_index_mapping = dict(sorted(inds.items()))
    return node_and_parent, node_to_solve_index_mapping


def bfs_edge_hops(graph: nx.DiGraph, root: Any, allowed_nodes_per_level: int):
    """Yields BFS tree edges along with hop count from root.

    Hop count increases when:
    - A real hop (child is at parent + 1 distance).
    - Or after `allowed_nodes_per_level` nodes have been discovered without hop
      increase.

    Function written by ChatGPT.

    Args:
        graph: The graph to traverse.
        root: The starting node for BFS.
        allowed_nodes_per_level: How many nodes are visited before the level is
            increased, even if the number of hops did not change.

    Yields:
        Tuple[Tuple[Any, Any], int]: Edge u, v and current hop count.
    """
    true_distance = nx.single_source_shortest_path_length(graph, root)

    current_depth = 0
    nodes_since_last_depth_increase = 0
    last_true_depth = true_distance[root]

    for u, v in nx.bfs_edges(graph, root):
        # Detect real hop
        if true_distance[v] > last_true_depth:
            current_depth += 1
            nodes_since_last_depth_increase = 0
            last_true_depth = true_distance[v]
        elif nodes_since_last_depth_increase >= allowed_nodes_per_level:
            current_depth += 1
            nodes_since_last_depth_increase = 0

        nodes_since_last_depth_increase += 1
        yield u, v, current_depth


def dhs_group_comps_into_levels(new_node_order: np.ndarray) -> List[np.ndarray]:
    """Group nodes into levels, such that nodes get processed in parallel when possible.

    Args:
        new_node_order: Array of shape (N, 3). The `3` are (node, parent, level).
            `N` is the number of compartment edges to be processed.

    Returns:
        Array of shape (num_levels, allowed_nodes_per_level, 2), where 2 indicates
        (node, parent).
    """
    if len(new_node_order) == 0:
        # If len == 0, we need to ensure that there are 3 columns, otherwise
        # pd.DataFrame() will throw an error.
        new_node_order = np.zeros((0, 3)).astype(int)

    # Group the edges by their level. Each level is processed in parallel.
    nodes = pd.DataFrame(new_node_order, columns=["node", "parent", "level"])
    grouping = nodes.groupby("level")
    nodes = grouping["node"].apply(list).to_numpy()
    parents = grouping["parent"].apply(list).to_numpy()
    nodes_and_parents = [np.stack([n, p]).T for n, p in zip(nodes, parents)]

    # `nodes_and_parents` is a List of length `num_levels`. Each element has shape
    # `(num_comps_per_level, 2)`.
    return nodes_and_parents



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from jaxley.__version__ import __version__
from jaxley.connect import (
    connect,
    connectivity_matrix_connect,
    fully_connect,
    sparse_connect,
)
from jaxley.integrate import integrate
from jaxley.io.swc import read_swc
from jaxley.modules import *
from jaxley.optimize import ParamTransform
from jaxley.stimulus import datapoint_to_step_currents, step_current

__all__ = [
    "read_swc",
    "Module",
    "Branch",
    "Cell",
    "Compartment",
    "Network",
]
